#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import arguments

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from beta_CROWN_solver import LiRPANet


def set_beta(self: 'LiRPANet', d, bias=True):
    # count how many split nodes in each batch example (batch, num of layers)
    splits_per_example = []
    max_splits_per_layer = {}
    batch = len(d['history'])
    for bi in range(batch):
        splits_per_example.append({})
        for k, v in d['history'][bi].items():
            # First element of layer_splits is a list of split neuron IDs.
            splits_per_example[bi][k] = len(v[0])
            max_splits_per_layer[k] = max(
                max_splits_per_layer.get(k, 0), splits_per_example[bi][k])

    # Create and load warmup beta.
    self.reset_beta(batch, max_splits_per_layer, betas=d['betas'], bias=bias)

    for node in self.split_nodes:
        if node.sparse_betas is None:
            continue
        sparse_betas = (node.sparse_betas
                        if isinstance(node.sparse_betas, list)
                        else node.sparse_betas.values())
        for sparse_beta in sparse_betas:
            sparse_beta.apply_splits(d['history'], node.name)

    return splits_per_example


def reset_beta(self: 'LiRPANet', batch, max_splits_per_layer, betas=None,
               bias=False):
    for layer_name in max_splits_per_layer:
        layer = self.net[layer_name]
        start_nodes = []
        for act in self.split_activations[layer_name]:
            start_nodes.extend(list(act[0].alpha.keys()))
        shape = (batch, max_splits_per_layer[layer_name])
        if betas is not None and betas[0] is not None and layer_name in betas[0]:
            betas_ = [(betas[bi][layer_name] if betas[bi] is not None else None)
                      for bi in range(batch)]
        else:
            betas_ = [None for _ in range(batch)]
        self.net.reset_beta(layer, shape, betas_, bias=bias,
                            start_nodes=list(set(start_nodes)))


def get_beta(self: 'LiRPANet', splits_per_example, device=None):
    # split_per_example only has half of the examples.
    beta_crown_args = arguments.Config["solver"]["beta-crown"]
    enable_opt_interm_bounds = beta_crown_args['enable_opt_interm_bounds']
    ret = []
    for i in range(len(splits_per_example)):
        betas = {}
        for k in splits_per_example[i]:
            if not enable_opt_interm_bounds:
                betas[k] = self._transfer(
                    self.net[k].sparse_betas[0].val[i, :splits_per_example[i][k]], device)
            else:
                betas[k] = []
                for sparse_beta in self.net[k].sparse_betas.values():
                    betas[k].append(self._transfer(
                        sparse_beta.val[i, :splits_per_example[i][k]], device))
        ret.append(betas)
    return ret



import math
import sys
import torch
import numpy as np
from auto_LiRPA import PerturbationLpNorm, BoundedTensor

from CTRAIN.bound import bound_ibp, bound_sabr
from CTRAIN.util import construct_c


def bound_taps(original_model, hardened_model, bounded_blocks, data, target, n_classes, ptb, device='cuda', pgd_steps=20, pgd_restarts=1, pgd_step_size=.2, 
               pgd_decay_factor=.2, pgd_decay_checkpoints=(5,7),
               gradient_link_thresh=.5, gradient_link_tolerance=1e-05, propagation="IBP", sabr_args=None):
    assert len(bounded_blocks) == 2, "Split not supported!"
    
    if propagation == 'IBP':
        lb, ub = bound_ibp(
            model=bounded_blocks[0],
            ptb=ptb,
            data=data,
            target=None,
            n_classes=n_classes,
        )
    if propagation == 'SABR':
        assert sabr_args is not None, "Need to Provide SABR arguments if you choose SABR for propagation"
        lb, ub = bound_sabr(
            # Intermediate Bound model instructs to return bounds after the first network block
            **{**sabr_args, "intermediate_bound_model": bounded_blocks[0], "return_adv_output": False},
        )
    
    with torch.no_grad():
        hardened_model.eval()
        original_model.eval()
        for block in bounded_blocks:
            block.eval()
        c = construct_c(data, target, n_classes)
        with torch.no_grad():
            grad_cleaner = torch.optim.SGD(hardened_model.parameters())
            adv_samples = _get_pivotal_points(bounded_blocks[1], lb, ub, pgd_steps, pgd_restarts, pgd_step_size, pgd_decay_factor, pgd_decay_checkpoints, n_classes, C=c)
            grad_cleaner.zero_grad()
            
        hardened_model.train()
        original_model.train()
        for block in bounded_blocks:
            block.train()
    
    pts = adv_samples[0].detach()
    pts = torch.transpose(pts, 0, 1)
    pts = RectifiedLinearGradientLink.apply(lb.unsqueeze(0), ub.unsqueeze(0), pts, gradient_link_thresh, gradient_link_tolerance)
    pts = torch.transpose(pts, 0, 1)
    pgd_bounds = _get_bound_estimation_from_pts(bounded_blocks[1], pts, None, c)
    # NOTE: VERY IMPORTANT CHANGES TO TAPS BOUND TO BE COMPATIBLE WITH CTRAIN WORKFLOW
    pgd_bounds = pgd_bounds[:, 1:]
    pgd_bounds = -pgd_bounds

        
    ibp_lb, ibp_ub = bound_ibp(
        model=bounded_blocks[1],
        ptb=PerturbationLpNorm(x_L=lb, x_U=ub),
        data=data,
        target=target,
        n_classes=n_classes,
    )

    return pgd_bounds, ibp_lb

# TODO: Adapted from TAPS code, should be checked if one wants to use multiestimator PGD loss alone
# TODO: Refactoring needed!!
def _get_pivotal_points(block, input_lb, input_ub, pgd_steps, pgd_restarts, pgd_step_size, pgd_decay_factor, pgd_decay_checkpoints, n_classes, C=None):
    '''
    This assumes the block net is fixed in this procedure. If a BatchNorm is involved, freeze its stat before calling this function.
    '''
    assert C is not None # Should only estimate for the final block
    lb, ub = input_lb.clone().detach(), input_ub.clone().detach()

    pt_list = []
    # split into batches
    # TODO: Can we keep this fixed batch size?
    bs = 128
    lb_batches = [lb[i*bs:(i+1)*bs] for i in range(math.ceil(len(lb) / bs))]
    ub_batches = [ub[i*bs:(i+1)*bs] for i in range(math.ceil(len(ub) / bs))]
    C_batches = [C[i*bs:(i+1)*bs] for i in range(math.ceil(len(C) / bs))]
    for lb_one_batch, ub_one_batch, C_one_batch in zip(lb_batches, ub_batches, C_batches):
        pt_list.append(_get_pivotal_points_one_batch(block, lb_one_batch, ub_one_batch, pgd_steps, pgd_restarts, pgd_step_size, pgd_decay_factor, pgd_decay_checkpoints, n_classes=n_classes, C=C_one_batch))
    pts = torch.cat(pt_list, dim=0)
    return [pts, ]

def _get_pivotal_points_one_batch(block, lb, ub, pgd_steps, pgd_restarts, pgd_step_size, pgd_decay_factor, pgd_decay_checkpoints, C, n_classes, device='cuda'):

    num_pivotal = n_classes - 1 # only need to estimate n_class - 1 dim for the final output

    def init_pts(input_lb, input_ub):
        rand_init = input_lb.unsqueeze(1) + (input_ub-input_lb).unsqueeze(1)*torch.rand(input_lb.shape[0], num_pivotal, *input_lb.shape[1:], device=device)
        return rand_init
    
    def select_schedule(num_steps):
        if num_steps >= 20 and num_steps <= 50:
            lr_decay_milestones = [int(num_steps*0.7)]
        elif num_steps > 50 and num_steps <= 80:
            lr_decay_milestones = [int(num_steps*0.4), int(num_steps*0.7)]
        elif num_steps > 80:
            lr_decay_milestones = [int(num_steps*0.3), int(num_steps*0.6), int(num_steps*0.8)]
        else:
            lr_decay_milestones = []
        return lr_decay_milestones

    lr_decay_milestones = pgd_decay_checkpoints
    lr_decay_factor = pgd_decay_factor
    init_lr = pgd_step_size

    retain_graph = False
    pts = init_pts(lb, ub)
    variety = (ub - lb).unsqueeze(1).detach()
    best_estimation = -1e5*torch.ones(pts.shape[:2], device=pts.device)
    best_pts = torch.zeros_like(pts)
    with torch.enable_grad():
        for re in range(pgd_restarts):
            lr = init_lr
            pts = init_pts(lb, ub)
            for it in range(pgd_steps+1):
                pts.requires_grad = True
                estimated_pseudo_bound = _get_bound_estimation_from_pts(block, pts, None, C=C)
                improve_idx = estimated_pseudo_bound[:, 1:] > best_estimation
                best_estimation[improve_idx] = estimated_pseudo_bound[:, 1:][improve_idx].detach()
                best_pts[improve_idx] = pts[improve_idx].detach()
                # wants to maximize the estimated bound
                if it != pgd_steps:
                    loss = - estimated_pseudo_bound.sum()
                    loss.backward(retain_graph=retain_graph)
                    new_pts = pts - pts.grad.sign() * lr * variety
                    pts = torch.max(torch.min(new_pts, ub.unsqueeze(1)), lb.unsqueeze(1)).detach()
                    if (it+1) in lr_decay_milestones:
                        lr *= lr_decay_factor
    return best_pts.detach()


def _get_bound_estimation_from_pts(block, pts, dim_to_estimate, C=None):
    '''
    only return estimated bounds for dims need to be estimated;
    '''
    if C is None:
        # pts shape (batch_size, num_pivotal, *shape_in[1:])
        out_pts = block(pts.reshape(-1, *pts.shape[2:]))
        out_pts = out_pts.reshape(*pts.shape[:2], -1)
        dim_to_estimate = dim_to_estimate.unsqueeze(1)
        dim_to_estimate = dim_to_estimate.expand(dim_to_estimate.shape[0], out_pts.shape[1], dim_to_estimate.shape[2])
        out_pts = torch.gather(out_pts, dim=2, index=dim_to_estimate) # shape: (batch_size, num_pivotal, num_pivotal)
        estimated_bounds = torch.diagonal(out_pts, dim1=1, dim2=2) # shape: (batch_size, num_pivotal)
    else:
        # # main idea: convert the 9 adv inputs into one batch to compute the bound at the same time; involve many reshaping
        batch_C = C.unsqueeze(1).expand(-1, pts.shape[1], -1, -1).reshape(-1, *(C.shape[1:])) # may need shape adjustment
        batch_pts = pts.reshape(-1, *(pts.shape[2:]))
        out_pts = block(batch_pts)
        out_pts = torch.bmm(batch_C, out_pts.unsqueeze(-1)).squeeze(-1)
        out_pts = out_pts.reshape(*(pts.shape[:2]), *(out_pts.shape[1:]))
        out_pts = - out_pts # the out is the lower bound of yt - yi, transform it to the upper bound of yi - yt
        # the out_pts should be in shape (batch_size, n_class - 1, n_class - 1)
        ub = torch.diagonal(out_pts, dim1=1, dim2=2) # shape: (batch_size, n_class - 1)
        estimated_bounds = torch.cat([torch.zeros(size=(ub.shape[0],1), dtype=ub.dtype, device=ub.device), ub], dim=1) # shape: (batch_size, n_class)

    return estimated_bounds

class RectifiedLinearGradientLink(torch.autograd.Function):
    r'''
    Estabilish Rectified linear gradient link between the input bounds and the input point.
    Note that this is not a valid gradient w.r.t. the forward function
    Take ub as an example: 
        For dims that x[dim] \in [lb, ub-c*(ub-lb)], the gradient w.r.t. ub is 0. 
        For dims that x[dim] == ub, the gradient w.r.t. ub is 1.
        For dims that x[dim] \in [ub-c*(ub-lb), ub], the gradient is linearly interpolated between 0 and 1.
    
    x should be modified to shape (batch_size, *bound_dims) by reshaping.
    bounds should be of shape (1, *bound_dims)
    '''
    @staticmethod
    def forward(ctx, lb, ub, x, c:float, tol:float):
        ctx.save_for_backward(lb, ub, x)
        ctx.c = c
        ctx.tol = tol
        return x
    @staticmethod
    def backward(ctx, grad_x):
        lb, ub, x = ctx.saved_tensors
        c, tol = ctx.c, ctx.tol
        slackness = c * (ub - lb)
        # handle grad w.r.t. ub
        thre = (ub - slackness)
        Rectifiedgrad_mask = (x >= thre)
        grad_ub = (Rectifiedgrad_mask * grad_x * (x - thre).clamp(min=0.5*tol) / slackness.clamp(min=tol)).sum(dim=0, keepdim=True)
        # handle grad w.r.t. lb
        thre = (lb + slackness)
        Rectifiedgrad_mask = (x <= thre)
        grad_lb = (Rectifiedgrad_mask * grad_x * (thre - x).clamp(min=0.5*tol) / slackness.clamp(min=tol)).sum(dim=0, keepdim=True)
        # we don't need grad w.r.t. x and param
        return grad_lb, grad_ub, None, None, None

class GradExpander(torch.autograd.Function):
    '''
    Multiply the gradient by alpha
    '''
    @staticmethod
    def forward(ctx, x, alpha:float=1):
        ctx.alpha = alpha
        return x
    
    @staticmethod
    def backward(ctx, grad_x):
        return ctx.alpha * grad_x, None


#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import torch
import numpy as np

def build_bound_implication_graph(cut_info):
    """
    Build the bound implication graph based on the extracted neuron implications.
    Graph node X corresponds to the actual neuron X // 2 in model. The parity of X indicates the neuron's inactive domain (even) or active domain (odd).

    dependency_edge_list: X => Y indicates the neuron X // 2's bound implicates neuron Y // 2's bound based on X or Y's parity.
    optimized_bound_indices: X => (Y, z) indicates with neuron X // 2's active/inactive bound, we can imply neuron Y's bound change to z.
    components[0]: X => {} stores all successor node in dependency graph given neuron X is assumed to be inactive (ub=0).
    components[1]: X => {} stores all successor node in dependency graph given neuron X is assumed to be active (lb=0).
    components[2]: X => {} stores all successor node in dependency graph based neurons X's split (i.e. union of components[0] and components[1]).
    """

    all_cuts = cut_info["all_cuts"]
    unstable_idx_list = cut_info["unstable_idx_list"]
    ref_idx = cut_info["ref_idx"]

    idx_mapping = {}
    all_unstable_neurons = 0
    for layers in range(len(unstable_idx_list)):
        for idx in (unstable_idx_list[layers]):
            idx_mapping[(ref_idx[layers], idx.item())] = all_unstable_neurons
            all_unstable_neurons += 1

    unstable_idx = []
    for layers in range(len(unstable_idx_list)):
        for idx in unstable_idx_list[layers]:
            unstable_idx.append([ref_idx[layers], idx.item()])

    total_layers = max(ref_idx) + 1
    optimized_bound_indices = [[] for i in range(all_unstable_neurons * 2 * total_layers * 2)]
    optimized_bound_values = [[] for i in range(all_unstable_neurons * 2 * total_layers * 2)]

    dependency_edge_list = [[] for i in range(all_unstable_neurons * 2)]

    all_improved = 0
    all_flipped = 0
    for key, _ in all_cuts.items():
        values = _["improved"]
        for i in range(values.shape[1]):
            nd, nidx = ref_idx[values[0][i].to(torch.long).item()], values[1][i].to(torch.long).item()
            base_idx = idx_mapping[(ref_idx[values[2][i].to(torch.long).item()], values[3][i].to(torch.long).item())]
            if key[-3] == "<": # base_idx's inactive domain
                base_idx = base_idx * 2
            else: # base_idx's active domain
                base_idx = base_idx * 2 + 1

            if key[0] == 'l': # target idx's lower bound
                optimized_bound_indices[base_idx * total_layers * 2 + nd * 2 + 0].append(nidx)
                optimized_bound_values[base_idx * total_layers * 2 + nd * 2 + 0].append(values[4][i].item())
            else:
                optimized_bound_indices[base_idx * total_layers * 2 + nd * 2 + 1].append(nidx)
                optimized_bound_values[base_idx * total_layers * 2 + nd * 2 + 1].append(values[4][i].item())

        all_improved += values.shape[1]

        values = _["flipped"]
        for i in range(values.shape[1]):
            improved_idx = idx_mapping[(ref_idx[values[0][i].to(torch.long).item()], values[1][i].to(torch.long).item())]
            base_idx = idx_mapping[(ref_idx[values[2][i].to(torch.long).item()], values[3][i].to(torch.long).item())]
            if key[-3] == "<": # base_idx's inactive domain
                base_idx = base_idx * 2
            else: # base_idx's active domain
                base_idx = base_idx * 2 + 1

            if key[0] == 'l': # target idx's lower bound > 0 => active domain
                improved_idx = improved_idx * 2 + 1
            else: # inactive domain otherwise
                improved_idx = improved_idx * 2
            dependency_edge_list[base_idx].append(improved_idx)
        all_flipped += values.shape[1]

    for i in range(all_unstable_neurons * 2 * total_layers * 2):
        optimized_bound_values[i] = torch.FloatTensor(optimized_bound_values[i]).cuda()
        optimized_bound_indices[i] = torch.LongTensor(optimized_bound_indices[i]).cuda()

    total_relus = total_layers

    dependency_statistics = np.zeros(total_relus).astype(np.int64) # the number of implications comes from each layer.
    indegree = np.zeros(all_unstable_neurons)
    dependency_statistics_all = np.zeros((total_relus, total_relus)) # the number of implications comes from layer to layer.
    for i in range(all_unstable_neurons * 2):
        dependencies = len(dependency_edge_list[i])
        dd = np.zeros(total_relus)
        for j in dependency_edge_list[i]:
            layer, idx = unstable_idx[j // 2]
            dd[layer] += 1
            indegree[j // 2] += 1
        layer, idx = unstable_idx[i // 2]
        dependency_statistics[layer] += dependencies
        for l in range(total_relus):
            dependency_statistics_all[layer][l] += dd[l]

    for _, i in enumerate(ref_idx):
        print(f"[Layer {i}]: All implications: {dependency_statistics[i]}")

    print("Implication matrix:")
    print(dependency_statistics_all)
    components = [[] for i in range(all_unstable_neurons)]

    max_comp_size, avg_comp_size = 0, 0

    def traverse(start_idx):
        node_queue = []
        node_queue.append(start_idx)
        visited = np.zeros(all_unstable_neurons * 2)
        visited[start_idx] = True
        head, tail = 0, 1
        while (head < tail):
            cur_node = node_queue[head]
            head += 1
            for neighbour in dependency_edge_list[cur_node]:
                if visited[neighbour] == False:
                    visited[neighbour] = True
                    node_queue.append(neighbour)
                    tail += 1
        return node_queue

    for neuron_idx in range(all_unstable_neurons):
        F_queue = traverse(neuron_idx * 2)
        T_queue = traverse(neuron_idx * 2 + 1)

        all_queue = []
        for i in range(len(F_queue)): all_queue.append(F_queue[i] // 2)
        for i in range(len(T_queue)): all_queue.append(T_queue[i] // 2)

        all_queue = np.unique(np.array(all_queue)).tolist()
        components[neuron_idx].append(F_queue)
        components[neuron_idx].append(T_queue)
        components[neuron_idx].append(all_queue)

        max_comp_size = max(max_comp_size, len(all_queue))
        avg_comp_size += len(all_queue)

    print("Total improved dependencies: %d, flipped dependencies: %d" % (all_improved, all_flipped))
    print("Maximum Comp size: %d, Average Comp size: %.4f" % (max_comp_size, avg_comp_size / all_unstable_neurons))

    return components, idx_mapping, optimized_bound_indices, optimized_bound_values, unstable_idx


import copy
import torch

from smac.utils.configspace import get_config_hash

from CTRAIN.model_wrappers.model_wrapper import CTRAINWrapper
from CTRAIN.train.certified.certified_training import sabr_train_model
from CTRAIN.util import seed_ctrain

class SABRModelWrapper(CTRAINWrapper):
    
    def __init__(self, model, input_shape, eps, num_epochs, train_eps_factor=1, optimizer_func=torch.optim.Adam, lr=0.0005, warm_up_epochs=1, ramp_up_epochs=70,
                 lr_decay_factor=.2, lr_decay_milestones=(80, 90), gradient_clip=10, l1_reg_weight=0.000001,
                 shi_reg_weight=.5, shi_reg_decay=True, sabr_subselection_ratio=.2, pgd_steps=8, 
                 pgd_alpha=0.5, pgd_restarts=1, pgd_early_stopping=False, pgd_alpha_decay_factor=.1,
                 pgd_decay_milestones=(4,7), checkpoint_save_path=None, 
                 bound_opts=dict(conv_mode='patches', relu='adaptive'), device=torch.device('cuda'),
                 ):
        super().__init__(model, eps, input_shape, train_eps_factor, lr, optimizer_func, bound_opts, device, checkpoint_save_path=checkpoint_save_path)
        self.cert_train_method = 'sabr'
        self.num_epochs = num_epochs
        self.lr = lr
        self.warm_up_epochs = warm_up_epochs
        self.ramp_up_epochs = ramp_up_epochs
        self.lr_decay_factor = lr_decay_factor
        self.lr_decay_milestones = lr_decay_milestones
        self.gradient_clip = gradient_clip
        self.l1_reg_weight = l1_reg_weight
        self.shi_reg_weight = shi_reg_weight
        self.shi_reg_decay = shi_reg_decay
        self.optimizer_func = optimizer_func
        self.sabr_subselection_ratio = sabr_subselection_ratio
        self.pgd_steps = pgd_steps
        self.pgd_alpha = pgd_alpha
        self.pgd_restarts = pgd_restarts
        self.pgd_early_stopping = pgd_early_stopping
        self.pgd_alpha_decay_factor = pgd_alpha_decay_factor
        self.pgd_decay_milestones = pgd_decay_milestones
        
        
    
    def train_model(self, train_loader, val_loader=None):
        eps_std = self.train_eps / train_loader.std if train_loader.normalised else torch.tensor(self.train_eps)

        eps_std = torch.reshape(eps_std, (*eps_std.shape, 1, 1))
        trained_model = sabr_train_model(
            original_model=self.original_model,
            hardened_model=self.bounded_model,
            train_loader=train_loader,
            val_loader=val_loader,
            num_epochs=self.num_epochs,
            eps=self.train_eps,
            eps_std=eps_std,
            eps_schedule=(self.warm_up_epochs, self.ramp_up_epochs),
            eps_scheduler_args={},
            optimizer=self.optimizer,
            lr_decay_schedule=self.lr_decay_milestones,
            lr_decay_factor=self.lr_decay_factor,
            n_classes=self.n_classes,
            gradient_clip=self.gradient_clip,
            l1_regularisation_weight=self.l1_reg_weight,
            shi_regularisation_weight=self.shi_reg_weight,
            shi_reg_decay=self.shi_reg_decay,
            subselection_ratio=self.sabr_subselection_ratio,
            pgd_steps=self.pgd_steps,
            pgd_step_size=self.pgd_alpha,
            pgd_restarts=self.pgd_restarts,
            pgd_early_stopping=self.pgd_early_stopping,
            pgd_decay_factor=self.pgd_alpha_decay_factor,
            pgd_decay_checkpoints=self.pgd_decay_milestones,
            results_path=self.checkpoint_path,
            device=self.device
        )
        
        return trained_model
    
    def _hpo_runner(self, config, seed, epochs, train_loader, val_loader, output_dir, cert_eval_samples=1000, include_nat_loss=True, include_adv_loss=True, include_cert_loss=True):
        config_hash = get_config_hash(config, 32)
        seed_ctrain(seed)
        
        if config['optimizer_func'] == 'adam':
            optimizer_func = torch.optim.Adam
        elif config['optimizer_func'] == 'radam':
            optimizer_func = torch.optim.RAdam
        if config['optimizer_func'] == 'adamw':
            optimizer_func = torch.optim.AdamW
        
        lr_decay_milestones = [
            config['warm_up_epochs'] + config['ramp_up_epochs'] + config['lr_decay_epoch_1'],
            config['warm_up_epochs'] + config['ramp_up_epochs'] + config['lr_decay_epoch_1'] + config['lr_decay_epoch_2']
        ]

        model_wrapper = SABRModelWrapper(
            model=copy.deepcopy(self.original_model), 
            input_shape=self.input_shape,
            eps=self.eps,
            num_epochs=epochs, 
            bound_opts=self.bound_opts,
            checkpoint_save_path=None,
            device=self.device,
            train_eps_factor=config['train_eps_factor'],
            optimizer_func=optimizer_func,
            lr=config['learning_rate'],
            warm_up_epochs=config['warm_up_epochs'],
            ramp_up_epochs=config['ramp_up_epochs'],
            gradient_clip=10,
            lr_decay_factor=config['lr_decay_factor'],
            lr_decay_milestones=[epoch for epoch in lr_decay_milestones if epoch <= epochs],
            l1_reg_weight=config['l1_reg_weight'],
            shi_reg_weight=config['shi_reg_weight'],
            shi_reg_decay=config['shi_reg_decay'],
            sabr_subselection_ratio=config['sabr:subselection_ratio'],
            pgd_alpha=config['sabr:pgd_alpha'],
            pgd_early_stopping=False,
            pgd_restarts=config['sabr:pgd_restarts'],
            pgd_steps=config['sabr:pgd_steps'],
            pgd_decay_milestones=()
        )

        model_wrapper.train_model(train_loader=train_loader)
        torch.save(model_wrapper.state_dict(), f'{output_dir}/nets/{config_hash}.pt')
        model_wrapper.eval()
        std_acc, cert_acc, adv_acc = model_wrapper.evaluate(test_loader=val_loader, test_samples=cert_eval_samples)

        loss = 0
        if include_nat_loss:
            loss -= std_acc
        if include_adv_loss:
            loss -= adv_acc
        if include_cert_loss:
            loss -= cert_acc

        return loss, {'nat_acc': std_acc, 'adv_acc': adv_acc, 'cert_acc': cert_acc} 



from CTRAIN.model_wrappers.shi_ibp_model_wrapper import ShiIBPModelWrapper
from CTRAIN.model_wrappers.crown_ibp_model_wrapper import CrownIBPModelWrapper
from CTRAIN.model_wrappers.sabr_model_wrapper import SABRModelWrapper
from CTRAIN.model_wrappers.taps_model_wrapper import TAPSModelWrapper
from CTRAIN.model_wrappers.staps_model_wrapper import STAPSModelWrapper
from CTRAIN.model_wrappers.mtl_ibp_model_wrapper import MTLIBPModelWrapper
from CTRAIN.model_wrappers.hpo_model_wrapper import HPOModelWrapper


#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
"""An simple example to run custom model and specs."""

import torch
import torch.nn as nn

class SimpleFeedForward(nn.Module):
    """A very simple model, just for demonstration."""
    def __init__(self, hidden_size):
        super().__init__()
        self.model = nn.Sequential(
            # Input dimension is 2, should match vnnlib file.
            nn.Linear(2, hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, hidden_size),
            nn.ReLU(),
            # Output dimension is 1, should match vnnlib file.
            nn.Linear(hidden_size, 1),
        )

    def forward(self, inputs):
        return self.model(inputs)

if __name__ == "__main__":
    # Save a random model for testing.
    model = SimpleFeedForward(hidden_size=32)
    torch.save(model.state_dict(), 'models/custom_specs/custom_specs.pth')




#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
"""Interface of cutting plane methods, using GCP-CROWN as the solver."""
import itertools
import torch
from auto_LiRPA.bound_ops import *


class Cutter:
    def __init__(self, solver, A=None, x=None, number_cuts=50, fix_intermediate_bounds=False, device='cuda'):
        self.solver = solver
        self.update_net(solver.net)
        # FIXME a['/input.1'] may not always be the input name
        self.A = [a['/input.1'] for a in A.values()] if A is not None else None
        self.x = x
        self.number_cuts = number_cuts
        self.fix_intermediate_bounds = fix_intermediate_bounds
        self.device = device
        self.unstable_idx_list, self.lower, self.upper = [], [], []
        self.lAs, self.uAs, self.lbs, self.ubs = [], [], [], []
        self.log_interval = 10
        self.beta_init = 0
        # Cutter should keep track of cuts
        self.cuts = []
        self.cut_module = None
        self.cut_timestamp = -1
        self.num_relus = len(self.relus)

    def update_net(self, net):
        self.net = net
        self.relus = net.relus

    def construct_cut_module(self, start_nodes=None, use_patches_cut=False):
        num_cuts = len(self.cuts)

        self.net.cut_used = False
        for m in self.relus:
            m.masked_beta = None
            # m.split_beta_used: True if any of this layer node is used in the new split constraints
            m.split_beta_used = False
            # m.history_beta_used: True if any of this layer node is used in the history constraints
            m.history_beta_used = False
            # if any of the current/history split constraints are single node constraint,
            m.single_beta_used = False
            # There are nodes used in cuts in this layer if cut_used=True
            m.cut_used = False
            m.relu_coeffs = m.arelu_coeffs = m.pre_coeffs = None

        self.start_nodes = start_nodes = [
            relu_layer.inputs[0] for relu_layer in self.net.relus] if not self.fix_intermediate_bounds else []
        start_nodes.append(self.net[self.net.output_name[0]])

        # init cut_module
        self.cut_module = cut_module = CutModule(self.relus)
        # maximum layer of the neuron in each constr
        # -1 is the input layer, 0 means the layer corresponding to the first relu layer
        max_layer_idx_in_constr = []
        cut_used = False
        self.pre_layers, self.relu_layers, self.arelu_layers = [], [], []
        self.use_x_cuts = False
        for cut_idx, ci in enumerate(self.cuts):
            # c = ci["c"] # 1: >0; -1: <0
            all_decisions = (ci["x_decision"] + ci["relu_decision"] +
                ci["arelu_decision"] + ci["pre_decision"])
            if len(all_decisions) > 0:
                max_layer_idx_in_constr.append(max([item[0] for item in all_decisions]))
            else:
                max_layer_idx_in_constr.append(-1)
            if len(ci["x_decision"]) > 0:
                cut_used = self.use_x_cuts = True
            self.relu_layers.extend([ self.relus[item[0]] for item in ci["relu_decision"]])
            self.arelu_layers.extend([ self.relus[item[0]] for item in ci["arelu_decision"]])
            self.pre_layers.extend([ self.relus[item[0]] for item in ci["pre_decision"]])
        self.pre_layers = set(self.pre_layers)
        self.relu_layers = set(self.relu_layers)
        self.arelu_layers = set(self.arelu_layers)
        for node in itertools.chain(self.relu_layers, self.arelu_layers, self.pre_layers):
            cut_used = node.cut_used = True
        if cut_used:
            self.net.cut_used = True

        # config active cuts
        print("all start nodes to check full crown or not:", start_nodes)
        print("use patches cut: ", use_patches_cut)
        cut_module.active_cuts = {}
        for node_idx, start_node in enumerate(start_nodes):
            active_cuts = []
            # use_patches_cut: disable to optimize intermediate layer that is in patches mode
            if not use_patches_cut and hasattr(start_node, "mode") and start_node.mode=="patches":
                print("skip cut beta crown opt for patches layer:", start_node)
                cut_module.active_cuts[start_node.name] = torch.tensor(active_cuts, device=self.device).long()
                continue
            if node_idx + 1 == len(start_nodes):
                cut_module.active_cuts[start_node.name] = torch.arange(num_cuts, device=self.device)
            else:
                for cut_idx in range(num_cuts):
                    # the constraints are active only if they have all constraint neurons before start node
                    # i.e., the max layer node in the constraint should be smaller than the current start node
                    if (max_layer_idx_in_constr[cut_idx] == -1 or
                            start_nodes.index(self.relus[max_layer_idx_in_constr[cut_idx]].inputs[0]) < node_idx):
                        active_cuts.append(cut_idx)
                cut_module.active_cuts[start_node.name] = torch.tensor(active_cuts, device=self.device).long()

        if self.use_x_cuts:
            assert self.x.size(0) == 1  # do not have batch dimension.
            self.cut_module.x_coeffs = torch.zeros((num_cuts, self.x.numel()), device=self.device)
        #FIXME maybe just use self.relus
        for node in self.relu_layers:
            cut_module.relu_coeffs[node.name] = torch.zeros((num_cuts, node.flattened_nodes), device=self.device)
        for node in self.arelu_layers:
            cut_module.arelu_coeffs[node.name] = torch.zeros((num_cuts, node.flattened_nodes), device=self.device)
        for node in self.pre_layers:
            cut_module.pre_coeffs[node.name] = torch.zeros((num_cuts, node.flattened_nodes), device=self.device)

        self.update_cut_module()

        return cut_module

    def construct_beta(self, shapes):
        general_betas = {}
        start_node_shape = {}
        for i, node in enumerate(self.net.relus):
            start_node_shape[node.inputs[0].name] = shapes[i]
        start_node_shape[self.net.final_name] = shapes[-1]
        # current all start nodes, only final_name if crown-ibp
        # all_start_nodes = self.net.backward_from[self.net.optimizable_activations[0].name]
        # manually get all start nodes, self.net.backward_from might not include first pre-relu layer
        # all_start_nodes = [relu_layer.inputs[0] for relu_layer in nodes] + [self.net[self.net.output_name[0]]]
        for start_node in self.start_nodes:
            # each general_beta: 2 (lA, uA), spec (out_c, out_h, out_w), batch, num_cuts
            general_betas[start_node.name] = self.beta_init * torch.ones(
                (2, *start_node_shape[start_node.name][1:], 1, len(self.cuts)), device=self.device)
            general_betas[start_node.name] = general_betas[start_node.name].detach()
            general_betas[start_node.name].requires_grad = True
        self.cut_module.general_beta = general_betas
        self.net.cut_beta_params = []
        for start_node in self.start_nodes:
            self.net.cut_beta_params.append(self.cut_module.general_beta[start_node.name])

    def init_cut(self, c=1):
        return {
            'x_decision': [], 'x_coeffs': [], 'arelu_decision': [], 'arelu_coeffs': [],
            'relu_decision': [], 'relu_coeffs': [], 'pre_decision': [], 'pre_coeffs': [],
            'c': c
        }

    def update_cut_module(self):
        # config the cut constraints coeffs in cut_module
        # add cut into each relu layer
        cut_bias = []
        for cut_idx, ci in enumerate(self.cuts):
            c = ci["c"] # 1: >0; -1: <0
            for node, coeff in zip(ci["x_decision"], ci["x_coeffs"]):
                layer, neuron_idx = node[0], node[1]
                self.cut_module.x_coeffs[cut_idx, neuron_idx] += c * coeff
            for node, coeff in zip(ci["relu_decision"], ci["relu_coeffs"]):
                layer, neuron_idx = node[0], node[1]
                self.cut_module.relu_coeffs[self.relus[layer].name][cut_idx, neuron_idx] += c * coeff
            for node, coeff in zip(ci["arelu_decision"], ci["arelu_coeffs"]):
                layer, neuron_idx = node[0], node[1]
                self.cut_module.arelu_coeffs[self.relus[layer].name][cut_idx, neuron_idx] += c * coeff
            for node, coeff in zip(ci["pre_decision"], ci["pre_coeffs"]):
                layer, neuron_idx = node[0], node[1]
                self.cut_module.pre_coeffs[self.relus[layer].name][cut_idx, neuron_idx] += c * coeff
            ##### have to be careful, c * bias (c=1) if constraint > bias;
            cut_bias.append(c * ci["bias"])
            #FIXME bias should be on torch
        self.cut_module.cut_bias = torch.tensor(cut_bias, device=self.device)

    def update_cuts(self):
        pass

    def refine_cuts(self):
        pass



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import copy
import time
import tqdm
import numpy as np

import torch

import arguments
from heuristics.base import NeuronBranchingHeuristic
from heuristics.nonlinear.babsr import BaBSRNonlinearBranching
from heuristics.nonlinear.utils import set_roots
from auto_LiRPA import BoundedTensor
from auto_LiRPA.bound_ops import *
from auto_LiRPA.utils import stop_criterion_batch_any, multi_spec_keep_func_all
from auto_LiRPA.utils import prod


class NonlinearBranching(NeuronBranchingHeuristic):
    """A general branching heuristic for nonlinear functions.

    TODO Scores are computed by calling the heuristic score function of the
    corresponding operator. We also want to implement optimizable branching
    point not limited to 0.
    """

    def __init__(self, net, **kwargs):
        super().__init__(net)
        self.input_split_method = 'sb'
        self.branching_point_method = kwargs.pop('branching_point_method')
        self.num_branches = kwargs.pop('num_branches')
        self.method = kwargs.pop('method')
        self.filter = kwargs.pop('filter')
        self.filter_beta = kwargs.pop('filter_beta')
        self.filter_batch_size = kwargs.pop('filter_batch_size')
        self.filter_iterations = kwargs.pop('filter_iterations')
        self.use_min = kwargs.pop('use_min')
        self.dynamic_bbps = kwargs.pop('dynamic_bbps')
        self.dynamic_options = kwargs.pop('dynamic_options')
        self.root_name = self.net.net.root_names[0]
        self.model = net.net
        self.roots = self.model.roots()
        if self.method == 'babsr-like':
            self.babsr = BaBSRNonlinearBranching(net, self.num_branches)


    def _get_uniform_branching_points(self, lb, ub, num_branches=None):
        if num_branches is None:
            num_branches = self.num_branches
        ratio = torch.arange(0, 1, step=1./num_branches)[1:].to(lb)
        assert ratio.shape[-1] == num_branches - 1
        points = lb.unsqueeze(-1) * (1 - ratio) + ub.unsqueeze(-1) * ratio
        return points

    def _get_input_split_scores(self, domains):
        print('Prioritizing input split for this round.')
        lb = domains['lower_bounds'][self.root_name]
        ub = domains['upper_bounds'][self.root_name]
        lA = domains['lAs'][self.root_name]
        if self.input_split_method == 'naive':
            scores = (ub - lb).flatten(1)
        elif self.input_split_method == 'sb':
            scores = (lA.abs() * (ub - lb).unsqueeze(1)).amax(dim=1)
        else:
            raise ValueError(self.input_split_method)
        scores = {self.root_name: scores}
        points = {
            self.root_name: self._get_uniform_branching_points(
                lb, ub).flatten(1, -2)
        }
        return scores, points

    def get_heuristic_decisions(self, domains, branching_candidates=1,
                                verbose=False, branching_point_method=None,
                                **kwargs):
        """Get rough decisions by a heuristic."""

        split_masks = domains['mask']
        self.update_batch_size_and_device(domains['lower_bounds'])

        scores = {}
        points = {}
        for node in self.net.split_nodes:
            name = node.name
            if verbose:
                print(f'Computing branching score for {name}')

            if self.dynamic_bbps:
                ret = self._dynamic_branching_points_bbps(
                    node, domains=domains)
            else:
                ret = self.compute_branching_scores(
                    node, domains=domains,
                    branching_point_method=branching_point_method)

            scores[name] = ret['scores'].flatten(1) * split_masks[node.name]
            scores[name] += split_masks[name] * 1e-10
            points[name] = ret['points'].flatten(1, -2)

        layers, indices, scores = self.find_topk_scores(
            scores, split_masks, branching_candidates, return_scores=True)
        num_branching_points = self.num_branches - 1
        points_ret = torch.full(
            (layers.shape[0], layers.shape[1], num_branching_points), -np.inf,
            device=scores.device)
        for idx, layer in enumerate(self.net.split_nodes):
            mask = layers.view(-1) == idx
            if mask.sum() == 0:
                continue
            name = layer.name
            indices_ = indices.clamp(max=points[name].shape[1]-1)
            points_ = torch.gather(
                points[name], dim=1,
                index=indices_.unsqueeze(-1).repeat(1, 1, num_branching_points))
            points_ret.view(-1, num_branching_points)[
                mask, :] = points_.view(-1, num_branching_points)[mask, :]

        points = points_ret

        return layers, indices, points

    def _dynamic_branching_points_bbps(self, node, domains):
        ret = None
        for i, bp in enumerate(self.dynamic_options):
            ret_ = self.compute_branching_scores(
                node, domains=domains, branching_point_method=bp)
            if ret is None:
                ret = ret_
                choice = torch.zeros_like(ret['scores']).to(torch.long)
            else:
                better = ret_['scores'] > ret['scores']
                choice[better] = i
                ret = {
                    'scores': torch.max(ret_['scores'], ret['scores']),
                    'points': torch.where(better.unsqueeze(-1),
                                          ret_['points'], ret['points']),
                }
        for i, bp in enumerate(self.dynamic_options):
            print(f'Choosing {bp}: {(choice == i).sum()}')
        return ret

    def get_branching_decisions(self, domains, split_depth=1,
                                branching_candidates=1, verbose=False, **kwargs):
        assert split_depth == 1
        if not self.filter:
            branching_candidates = 1

        layers, indices, points = self.get_heuristic_decisions(
            domains, branching_candidates=branching_candidates,
            verbose=verbose, **kwargs)
        if self.filter:
            layers, indices, points = self._filter(
                domains, layers, indices, points)

        return self.format_decisions(layers, indices, points)

    def _compute_actual_bounds(self, domains, decisions, iterations=None):
        if iterations is None:
            iterations = self.filter_iterations

        lb, ub = domains['lower_bounds'], domains['upper_bounds']
        args_update_bounds = {
            'lower_bounds': lb, 'upper_bounds': ub,
            'alphas': domains['alphas'], 'cs': domains['cs'],
            'thresholds': domains.get('thresholds', None)
        }
        if self.filter_beta:
            args_update_bounds.update({
                'betas': domains['betas'],
                'history': domains['history']
            })
        print('Start filtering...')
        branching_decision, branching_points, _ = decisions
        split = {
            'decision': branching_decision,
            'points': branching_points,
        }
        self.net.build_history_and_set_bounds(
            args_update_bounds, split, mode='breath')

        total_candidates = args_update_bounds['cs'].shape[0]
        num_batches = (
            total_candidates + self.filter_batch_size - 1
        ) // self.filter_batch_size
        ret_lbs = []
        iterations_bak = arguments.Config['solver']['beta-crown']['iteration']
        arguments.Config['solver']['beta-crown']['iteration'] = iterations
        for i in tqdm.tqdm(range(num_batches)):
            args_update_bounds_ = self._take_filter_batch(args_update_bounds, i)
            ret_lbs_ = self.net.update_bounds(
                copy.deepcopy(args_update_bounds_),
                shortcut=True, beta=self.filter_beta, beta_bias=True,
                stop_criterion_func=stop_criterion_batch_any(args_update_bounds_['thresholds']),
                multi_spec_keep_func=multi_spec_keep_func_all)
            ret_lbs.append(ret_lbs_.detach())
        arguments.Config['solver']['beta-crown']['iteration'] = iterations_bak
        ret_lbs = torch.concat(ret_lbs, dim=0)
        return ret_lbs

    def _take_filter_batch(self, args_update_bounds, i):
        args_update_bounds_ = {
            'lower_bounds': {
                k: v[i*self.filter_batch_size:(i+1)*self.filter_batch_size]
                for k, v in args_update_bounds['lower_bounds'].items()},
            'upper_bounds': {
                k: v[i*self.filter_batch_size:(i+1)*self.filter_batch_size]
                for k, v in args_update_bounds['upper_bounds'].items()}
        }
        for k in ['cs', 'thresholds']:
            args_update_bounds_[k] = args_update_bounds[k][
                i*self.filter_batch_size:(i+1)*self.filter_batch_size]
        if self.filter_beta:
            for k in ['betas', 'history']:
                args_update_bounds_[k] = (args_update_bounds[k][
                    i*self.filter_batch_size:(i+1)*self.filter_batch_size])
        args_update_bounds_['alphas'] = {
            k: {kk: vv[:, :, i*self.filter_batch_size:(i+1)*self.filter_batch_size]
                for kk, vv in v.items()}
            for k, v in args_update_bounds['alphas'].items()
        }
        return args_update_bounds_

    def _filter(self, domains, layers, indices, points, verbose=False):
        filter_start_time = time.time()
        decisions = self.format_decisions(layers, indices, points)

        ret_lbs = self._compute_actual_bounds(domains, decisions)

        kfsb_scores = []
        # [branching_candidates, num_branches, num_domains]
        branching_candidates = layers.shape[1]
        scores = (
            ret_lbs.view(branching_candidates, self.num_branches,
                         -1, ret_lbs.shape[-1]) - domains['thresholds']
        ).amax(dim=-1)
        kfsb_scores = scores.mean(dim=1)
        kfsb_choice = kfsb_scores.argmax(dim=0)

        print('kfsb scores (first 10):', kfsb_scores[:, :10])
        print('kfsb choice (first 10):', kfsb_choice[:10])

        layers = torch.gather(layers, index=kfsb_choice.unsqueeze(-1), dim=1)
        indices = torch.gather(indices, index=kfsb_choice.unsqueeze(-1), dim=1)
        kfsb_choice = kfsb_choice.unsqueeze(-1).unsqueeze(-1).repeat(
            1, 1, self.num_branches - 1)
        points = torch.gather(points, index=kfsb_choice, dim=1)
        lb_old = domains['lower_bounds'][self.net.final_name]
        if verbose:
            previous_best = (lb_old-domains['thresholds']).amax(dim=-1).max()
            previous_worst = (lb_old-domains['thresholds']).amax(dim=-1).min()
            new_worst = kfsb_scores.max(dim=0).values.min()
            print('Previous best:', previous_best)
            print('Previous worst:', previous_worst)
            print('New worst:', new_worst)
        print('Filtering time:', time.time() - filter_start_time)
        return layers, indices, points

    def compute_branching_scores(self, node, domains, branching_point_method=None):
        lb_ori = domains['lower_bounds'][node.name]
        ub_ori = domains['upper_bounds'][node.name]

        if branching_point_method is None:
            branching_point_method = self.branching_point_method

        if branching_point_method == 'uniform':
            points = self._get_uniform_branching_points(
                lb_ori, ub_ori, self.num_branches)
        elif branching_point_method in ['three_left', 'three_right']:
            points = self._get_uniform_branching_points(lb_ori, ub_ori, 3)
            if branching_point_method == 'three_left':
                points = points[..., :1]
            else:
                points = points[..., -1:]
        else:
            raise NameError(branching_point_method)

        return self.compute_scores_with_points(node, domains, points)

    def compute_scores_with_points(self, node, domains, points):
        name = node.name
        lb, ub = domains['lower_bounds'], domains['upper_bounds']
        lb_ori, ub_ori = lb[name], ub[name]

        for n in self.net.net.nodes():
            if hasattr(n, 'opt_stage'):
                n.opt_stage = None

        start_nodes = [act[0] for act in self.net.split_activations[name]]

        # Special cases for now
        if len(start_nodes) == 1:
            if isinstance(start_nodes[0], (BoundRelu, BoundSign, BoundSignMerge)):
                # For ReLU or LeakyReLU, always branch at 0.
                mask_unstable = torch.logical_and(lb_ori < 0, ub_ori > 0)
                points[mask_unstable, :] = 0

        global_lb = domains['lower_bounds'][self.net.final_name]
        margin_before = (global_lb - domains['thresholds']).unsqueeze(1)

        if self.method == 'babsr-like':
            return self.babsr.compute_heuristic(
                node, points, domains, margin_before)
        else:
            return self._fast_heuristic(
                node, points, domains, margin_before)

    def _fast_heuristic(self, node, points, domains, margin_before):
        name = node.name
        lAs = domains['lAs']
        lb, ub = domains['lower_bounds'], domains['upper_bounds']
        lb_ori, ub_ori = lb[name], ub[name]
        start_nodes = [act[0] for act in self.net.split_activations[name]]

        A_before, bound_before, unstable_idx = self._fast_backward_propagation(
            lAs, lb, ub, node, start_nodes)
        dim_output, batch_size, num_neurons = bound_before.shape
        assert isinstance(self.roots[0], BoundInput)
        assert isinstance(self.roots[0].value, BoundedTensor)
        x_new = self.net.expand_x(batch_size)
        set_roots(self.roots, x_new, A_before)
        # (batch_size, dim_output)
        bound_from_A = self.model.concretize(
            batch_size, dim_output,
            torch.zeros((batch_size, dim_output), device=bound_before.device),
            bound_upper=False)[0]
        # (batch_size, num_neurons, dim_output)
        bound_before = bound_from_A.unsqueeze(1) + bound_before.permute(1, 2, 0)

        margin_afters = []
        for i in range(self.num_branches):
            lb_branched = lb_ori if i == 0 else points[..., i - 1]
            ub_branched = ub_ori if i + 1 == self.num_branches else points[..., i]

            lb_ = {k: lb_branched if k == name else v
                   for k, v in lb.items()}
            ub_ = {k: ub_branched if k == name else v
                   for k, v in ub.items()}

            A_after, bound_after, _ = self._fast_backward_propagation(
                lAs, lb_, ub_, node, start_nodes)
            # A_before: (dim_output, batch_size, num_neurons, dim_input)
            diff_A = A_after - A_before
            A_ = A_before.sum(dim=2, keepdim=True) + diff_A
            # (dim_output * num_neurons, batch_size, dim_input)
            self.roots[0].lA = A_.transpose(1, 2).reshape(
                -1, batch_size, A_after.shape[-1])
            # (batch_size, dim_output * num_neurons)
            bound_after = bound_after.transpose(
                1, 2).reshape(-1, batch_size).transpose(0, 1)
            # (batch_size, dim_output, num_neurons)
            bound_after = self.model.concretize(
                bound_after.shape[0], bound_after.shape[1],
                bound_after, bound_upper=False)[0]
            # (batch_size, num_neurons, dim_output)
            bound_after = bound_after.reshape(
                batch_size, dim_output, num_neurons).transpose(1, 2)
            # (batch_size, dim_output, num_neurons)
            bound_delta = bound_after - bound_before
            margin_after_ = margin_before + bound_delta
            margin_afters.append(margin_after_)

        margin_after = torch.concat(margin_afters).reshape(
            -1, *margin_afters[0].shape)

        margin_after_min = margin_after.clone()
        margin_after_sum = margin_after.clone()
        margin_after_min = margin_after_min.min(dim=0).values
        margin_after_sum = margin_after_sum.sum(dim=0)

        def _unfold_margin_after(margin, unstable_idx):
            margin_full = torch.zeros(
                margin.shape[0], *node.output_shape[1:],
                margin.shape[-1], device=margin.device)
            if isinstance(unstable_idx, torch.Tensor):
                margin_full[:, unstable_idx, :] = margin
            else:
                margin_full[:, unstable_idx[0],
                            unstable_idx[1], unstable_idx[2], :] = margin
            return margin_full.reshape(
                margin_full.shape[0], -1,  margin_full.shape[-1])

        if unstable_idx is not None:
            margin_after_min = _unfold_margin_after(
                margin_after_min, unstable_idx)
            margin_after_sum = _unfold_margin_after(
                margin_after_sum, unstable_idx)

        ret = {
            'points': points.flatten(1, -2),
            'margin_after': margin_after,
        }

        scores_min = (margin_after_min - margin_before).amin(dim=-1)
        scores_sum = (margin_after_sum - margin_before).sum(dim=-1)

        if self.use_min:
            ret['scores'] = scores_min
        else:
            ret['scores'] = scores_sum

        return ret

    def _fast_backward_propagation(self, lAs, lb, ub, branched_node, start_nodes):
        A_root = None
        bound = None
        unstable_idx = None
        for node in start_nodes:
            lA = lAs[node.name].transpose(0, 1)
            for i in node.requires_input_bounds:
                inp = node.inputs[i]
                inp.lower, inp.upper = lb[inp.name], ub[inp.name]
            if not isinstance(node, (BoundActivation,
                                     BoundOptimizableActivation)):
                raise NotImplementedError
            with torch.no_grad():
                A, lower_b, _ = node.bound_backward(
                    lA, None, *node.inputs,
                    start_node=self.model[self.model.final_name], reduce_bias=False)
            for i, node_pre in enumerate(node.inputs):
                if node_pre != branched_node:
                    continue
                bound_ = lower_b[i] if isinstance(lower_b, tuple) else lower_b
                if bound_.shape[2:] != branched_node.output_shape[1:]:
                    print('Error: incorrect shapes in the branching heuristic.')
                    print('It may be because that _fast_backward_propagation has '
                          f'not been supported for {node} yet')
                    print('Please debug:')
                    import pdb; pdb.set_trace()

                def maybe_convert_A(A):
                    if isinstance(A, torch.Tensor):
                        return A
                    else:
                        return A.to_matrix(self.roots()[0].output_shape)

                A_saved = self.net.A_saved[node_pre.name][self.model.input_name[0]]
                lA_next = maybe_convert_A(A_saved['lA'])
                assert lA_next.shape[0] == 1
                lA_next = lA_next.reshape(lA_next.shape[1], -1)
                uA_next = maybe_convert_A(A_saved['uA'])
                assert uA_next.shape[0] == 1
                uA_next = uA_next.reshape(uA_next.shape[1], -1)
                lbias = A_saved['lbias']
                ubias = A_saved['ubias']
                lbias = lbias.flatten() if lbias is not None else None
                ubias = ubias.flatten() if ubias is not None else None
                A_ = A[i][0]

                if A_saved['unstable_idx'] is not None:
                    assert unstable_idx is None
                    unstable_idx = A_saved['unstable_idx']
                    if isinstance(A_saved['unstable_idx'], torch.Tensor):
                        A_ = A_[:, :, A_saved['unstable_idx']]
                        assert A_saved['unstable_idx'].max()<bound_.shape[2]
                        bound_ = bound_[:, :, A_saved['unstable_idx']]
                    else:
                        A_ = A_[:, :,
                                A_saved['unstable_idx'][0],
                                A_saved['unstable_idx'][1],
                                A_saved['unstable_idx'][2]]
                        assert A_saved['unstable_idx'][0].max()<bound_.shape[2]
                        assert A_saved['unstable_idx'][1].max()<bound_.shape[3]
                        assert A_saved['unstable_idx'][2].max()<bound_.shape[4]
                        bound_ = bound_[:, :, A_saved['unstable_idx'][0],
                                        A_saved['unstable_idx'][1],
                                        A_saved['unstable_idx'][2]]

                A_ = A_.reshape(A_.shape[0], A_.shape[1], -1)
                bound_ = bound_.reshape(bound_.shape[0], bound_.shape[1], -1)

                if lbias is not None:
                    bound_  = bound_ + (A_.clamp(min=0) * lbias
                                        + A_.clamp(max=0) * ubias)

                A_ = A_.unsqueeze(-1)
                A_root_ = A_.clamp(min=0) * lA_next + A_.clamp(max=0) * uA_next
                if A_root is None:
                    A_root = A_root_
                else:
                    assert A_root.shape == A_root_.shape
                    A_root += A_root_

                if bound is None:
                    bound = bound_
                else:
                    bound += bound_

        if bound.ndim < 3:
            breakpoint()
        return A_root, bound, unstable_idx




#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
"""
Old branching heuristics, must be removed very soon (assigned to Kaidi).
"""

import torch
from auto_LiRPA import BoundedTensor, PerturbationLpNorm
import arguments


@torch.no_grad()
def input_split_branching(net, dom_lb, x_L, x_U, lA, thresholds,
                          branching_method, split_depth=1, num_iter=0,
                          last_split_idx=None):
    """
    Produce input split according to branching methods.
    """
    x_L = x_L.flatten(1)
    x_U = x_U.flatten(1)

    if branching_method == 'naive':
        # we just select the longest edge
        return torch.topk(x_U - x_L, split_depth, -1).indices
    elif branching_method == 'sb':
        return input_split_heuristic_sb(
            x_L, x_U, dom_lb, thresholds, lA, split_depth)
    elif branching_method == 'brute-force':
        assert split_depth == 1
        if num_iter <= arguments.Config['bab']['branching']['input_split']['bf_iters']:
            return input_split_heuristic_bf(
                net, x_L, x_U, dom_lb, thresholds, lA, last_split_idx)
        else:
            return input_split_heuristic_sb(
                x_L, x_U, dom_lb, thresholds, lA, split_depth)
    else:
        raise NameError(f'Unsupported branching method "{branching_method}" for input splits.')


def input_split_heuristic_sb(x_L, x_U, dom_lb, thresholds, lA, split_depth=1):
    branching_args = arguments.Config['bab']['branching']
    input_split_args = branching_args['input_split']
    lA_clamping_thresh = input_split_args['sb_coeff_thresh']
    sb_margin_weight = input_split_args['sb_margin_weight']
    sb_sum = input_split_args['sb_sum']
    touch_zero_score = input_split_args['touch_zero_score']

    lA = lA.view(lA.shape[0], lA.shape[1], -1)
    # lA shape: (batch, spec, # inputs)
    perturb = (x_U - x_L).unsqueeze(-2)
    # perturb shape: (batch, 1, # inputs)
    # dom_lb shape: (batch, spec)
    # thresholds shape: (batch, spec)
    assert lA_clamping_thresh >= 0

    if sb_sum:
        score = lA.abs().clamp(min=lA_clamping_thresh) * perturb / 2
        score = score.sum(dim=-2)
        if touch_zero_score:
            touch_zero = torch.logical_or(x_L == 0, x_U == 0)
            score = score + touch_zero * (x_U - x_L) * touch_zero_score
    else:
        score = (lA.abs().clamp(min=lA_clamping_thresh) * perturb / 2
                + (dom_lb.to(lA.device).unsqueeze(-1)
                    - thresholds.unsqueeze(-1)) * sb_margin_weight)
        score = score.amax(dim=-2)
    # note: the k (split_depth) in topk <= # inputs, because split_depth is computed as
    # min(max split depth, # inputs).
    # 1) If max split depth <= # inputs, then split_depth <= # inputs.
    # 2) If max split depth > # inputs, then split_depth = # inputs.
    return torch.topk(score, split_depth, -1).indices


def input_split_heuristic_bf(net, x_L, x_U, dom_lb, thresholds, lA,
                             last_split_idx):
    branching_args = arguments.Config['bab']['branching']
    input_split_args = branching_args['input_split']
    lA_clamping_thresh = input_split_args['sb_coeff_thresh']
    sb_margin_weight = input_split_args['sb_margin_weight']
    bf_backup_thresh = input_split_args['bf_backup_thresh']
    bf_rhs_offset = input_split_args['bf_rhs_offset']
    zero_crossing_score = input_split_args['bf_zero_crossing_score']
    touch_zero_score = input_split_args['touch_zero_score']

    assert x_L.ndim == 2
    input_dim = x_L.shape[1]
    x_M = (x_L + x_U) / 2
    new_x_L = x_L.expand(2, input_dim, -1, -1).clone()
    new_x_U = x_U.expand(2, input_dim, -1, -1).clone()
    for i in range(input_dim):
        new_x_U[0, i, :, i] = x_M[:, i]
        new_x_L[1, i, :, i] = x_M[:, i]
    new_x_L = new_x_L.view(-1, new_x_L.shape[-1])
    new_x_U = new_x_U.view(-1, new_x_U.shape[-1])
    new_x = BoundedTensor(
        new_x_L,
        ptb=PerturbationLpNorm(x_L=new_x_L, x_U=new_x_U))
    C = net.c.expand(new_x.shape[0], -1, -1)
    lb_ibp = net.net.compute_bounds(
        x=(new_x,), C=C, method='ibp', bound_upper=False)[0]
    reference_interm_bounds = {}
    for node in net.net.nodes():
        if (node.perturbed
                and isinstance(node.lower, torch.Tensor)
                and isinstance(node.upper, torch.Tensor)):
            reference_interm_bounds[node.name] = (node.lower, node.upper)
    lb_crown = net.net.compute_bounds(
        x=(new_x,), C=C, method='crown', bound_upper=False,
        reference_bounds=reference_interm_bounds
    )[0]
    lb = torch.max(lb_ibp, lb_crown)

    margin = (lb - thresholds[0]).view(2, input_dim, -1, lb.shape[-1])
    lb_base = dom_lb.cuda() - thresholds[0]
    verified = margin.amax(dim=-1) > 0

    assert bf_rhs_offset >= 0
    objective = (
        (margin - lb_base).clamp(min=0)
        / (lb_base - bf_rhs_offset).abs().clamp(min=1e-8)
        * (1 - verified.unsqueeze(-1).int())
    ).clamp(max=2e8).sum(dim=0)

    objective = objective.sum(dim=-1)
    objective = objective + 1e9 * verified.sum(dim=0)
    too_bad = objective.amax(dim=0) < bf_backup_thresh

    # TODO branch at zero rather than midpoint
    if zero_crossing_score:
        cross_zero = torch.logical_and(x_L < 0, x_U > 0)
        objective = objective + (cross_zero * (x_U - x_L) * 50000).t()
    if touch_zero_score:
        touch_zero = torch.logical_or(x_L == 0, x_U == 0)
        objective = objective + (touch_zero * (x_U - x_L) * touch_zero_score).t()

    lA = lA.view(lA.shape[0], lA.shape[1], -1)
    perturb = (x_U - x_L).unsqueeze(-2)
    sb_score = (lA.abs().clamp(min=lA_clamping_thresh) * perturb / 2
            + (dom_lb.to(lA.device).unsqueeze(-1)
                - thresholds.unsqueeze(-1)) * sb_margin_weight)
    sb_score = sb_score.sum(dim=-2)
    objective[:, too_bad] = sb_score[too_bad].t()

    index = objective.argmax(0).unsqueeze(-1)

    worst_idx = margin.amax(dim=-1).amin(dim=0).amax(dim=0).argmin()
    print('Worst idx:', worst_idx)
    print('Before', lb_base[worst_idx])
    print('Left branch:', margin[0, :, worst_idx])
    print('Right branch:', margin[1, :, worst_idx])
    print('Selected index:', index[worst_idx])
    if last_split_idx is not None:
        print('Last selected index:', last_split_idx[worst_idx])
    print('Objective', objective[:, worst_idx])
    print('x_L', x_L[worst_idx])
    print('x_U', x_U[worst_idx])
    if too_bad[worst_idx]:
        print('Bad objective. Using SB.')

    if torch.isnan(margin).any():
        import pdb; pdb.set_trace()

    return index



import torch
import torch.optim as optim
import torch.nn as nn

from art.estimators.classification import PyTorchClassifier
from art.attacks.evasion import ProjectedGradientDescent

from CTRAIN.eval import eval_epoch
from CTRAIN.util import save_checkpoint

def pgd_train_model(hardened_model, train_loader, val_loader=None, num_epochs=100, eps=0.3, input_shape=(1, 28, 28), n_classes=10, learning_rate=0.001,
                    results_path="./results"):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    optimizer=optim.Adam(lr=learning_rate, params=hardened_model.parameters())
    criterion = nn.CrossEntropyLoss()
    classifier = PyTorchClassifier(
        model=hardened_model,
        clip_values=(-0.42421296, 2.82148671),
        input_shape=(1, 28, 28),
        nb_classes=n_classes,
        loss=criterion
    )
    attack = ProjectedGradientDescent(
        estimator=classifier,
        eps=eps,
        eps_step=eps/4.0,
        max_iter=7,
        num_random_init=1,
        verbose=False
    )
    
    for epoch in range(num_epochs):
        hardened_model.train()
        running_loss = 0.0
        for data, target in train_loader:
            data, target = data.to(device), target.to(device)
            adv_samples = attack.generate(data.cpu().numpy())
            adv_samples = torch.from_numpy(adv_samples).to(device)
            optimizer.zero_grad()
            output = hardened_model(adv_samples)
            loss = criterion(output, target)
            loss.backward()
            optimizer.step()
            
            running_loss += loss.item()
        
        print(f'Epoch [{epoch+1}/{num_epochs}], Loss: {running_loss/len(train_loader):.4f}')
        
        correct_clean, correct_adv, total = 0, 0, 0
        for data, target in val_loader:
            data, target = data.to(device), target.to(device)
            output = hardened_model(data)
            predicted = torch.argmax(output, dim=1)
            total += target.size(0)
            correct_clean += torch.sum(predicted == target)
            adv_samples = attack.generate(data.cpu().numpy())
            adv_samples = torch.from_numpy(adv_samples).to(device)
            adv_output = hardened_model(adv_samples)
            adv_predicted = torch.argmax(adv_output, dim=1)
            correct_adv += torch.sum(adv_predicted == target)
            
        print(f"Epoch [{epoch+1}/{num_epochs}], VAL ACC: {correct_clean / total}, ADV. VAL. ACC: {correct_adv / total}")
        eval_epoch(
            model=hardened_model,
            optimizer=optimizer,
            loss=running_loss,
            data_loader=train_loader,
            eps=eps,
            n_classes=n_classes,
            epoch=epoch,
            device=device,
            results_path=f"{results_path}/{epoch}"
        )
        save_checkpoint(hardened_model, optimizer, running_loss, epoch + 1, f"{results_path}/{epoch}")

    return hardened_model




#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import math
import time
import torch
import torch.nn as nn
from torch.optim import Optimizer
import numpy as np
from tqdm import tqdm
import arguments
import os
import subprocess
from load_model import Customized
from attack.attack_pgd import default_pgd_loss
import sys

def customized_gtrsb_loss(origin_out, output, C_mat, rhs_mat, cond_mat, same_number_const,
               gama_lambda=0, threshold=-1e-5, mode='hinge', model=None):
    # first return the original loss
    loss, loss_gama = default_pgd_loss(origin_out, output, C_mat, rhs_mat, cond_mat, same_number_const,
               gama_lambda, threshold, mode)
    signmerge_loss = torch.zeros_like(loss, device=loss.device)
    signmerge_losses = []

    _, num_restarts, num_specs, _ = output.shape

    # pick up all SignMerge layers from the model
    model_layers = model._modules.keys()
    signmerge_layers = []
    for layer_id in model_layers:
        if "/merge" in layer_id:
            signmerge_layers.append(layer_id)

    threshold = 1e-4
    scaler = 10

    for layer_id in signmerge_layers[1:]:   # ignore the first SignMerge layer since there isn't any error
        input_signmerge = model.get_forward_value(model[layer_id].inputs[0])    # the input of this SignMerge layer
        elementwise_loss = torch.clamp(threshold - torch.abs(input_signmerge), min=0)
        elementwise_loss = elementwise_loss.view(num_restarts * num_specs, -1)
        layer_loss = torch.mean(elementwise_loss, dim=1)
        signmerge_losses.append(-layer_loss.view(loss.shape))

    signmerge_loss = torch.mean(torch.stack(signmerge_losses), dim=0) / (scaler * threshold)

    loss_gama += torch.sum(signmerge_loss)

    return loss, loss_gama




import torch
from auto_LiRPA import BoundedTensor

from CTRAIN.util import construct_c

# Uses CROWN bounds throughout all intermediate layers and the final layer. 
def bound_crown(model, ptb, data, target, n_classes=10, bound_upper=False, reuse_input=False):
    data = BoundedTensor(data, ptb=ptb)
    c = construct_c(data, target, n_classes)
    if reuse_input:
        bound_input = None
    else:
        bound_input = (data,)
    lb, ub = model.compute_bounds(x=bound_input, IBP=False, method="CROWN", C=c, bound_upper=bound_upper)
    return lb, ub

# CROWN-IBP uses IBP bounds for all intermediate layers and CROWN bounds for the last one
def bound_crown_ibp(model, ptb, data, target, n_classes=10, bound_upper=False, reuse_input=False):
    data = BoundedTensor(data, ptb=ptb)
    c = construct_c(data, target, n_classes)
    if reuse_input:
        bound_input = None
    else:
        bound_input = (data,)
    lb, ub = model.compute_bounds(x=bound_input, IBP=False, method="CROWN-IBP", C=c, bound_upper=bound_upper)
    return lb, ub


import torch
import torch.nn as nn


def get_l1_reg(model, device='cuda'):
    loss = torch.zeros(()).to(device)
    # only regularise Linear and Convolutional layers
    for module in model.modules():
        if isinstance(module, nn.Linear):
            loss += torch.abs(module.weight).sum()
        elif isinstance(module, nn.Conv2d):
            loss += torch.abs(module.weight).sum()
    return loss


#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
"""Various kinds of specifications for verification."""

import arguments
import torch
import numpy as np

from beta_CROWN_solver import LiRPANet

class Specification:
    def __init__(self):
        self.num_outputs = arguments.Config['data']['num_outputs']
        # FIXME Do not use numpy. Use torch instead.
        self.rhs = np.array([arguments.Config['bab']['decision_thresh']])

    def construct_vnnlib(self):
        raise NotImplementedError


class SpecificationVerifiedAcc(Specification):
    def construct_vnnlib(self, dataset, x_range, example_idx_list):
        vnnlib = []
        for i in range(len(example_idx_list)):
            label = dataset['labels'][example_idx_list[i]].view(1, 1)
            this_x_range = x_range[i]
            c = (torch.eye(self.num_outputs)[label].unsqueeze(1)
                - torch.eye(self.num_outputs).unsqueeze(0))
            I = (~(label.unsqueeze(1) == torch.arange(
                    self.num_outputs).type_as(label.data).unsqueeze(0)))
            c = c[I].view(1, self.num_outputs - 1, self.num_outputs)
            new_c = []
            for ii in range(self.num_outputs - 1):
                new_c.append((c[:, ii], self.rhs))
            vnnlib.append([(this_x_range, new_c)])
        return vnnlib


class SpecificationTarget(Specification):
    def construct_vnnlib(self, dataset, x_range, example_idx_list):
        vnnlib = []
        for i in range(len(example_idx_list)):
            label = dataset['labels'][example_idx_list[i]].view(1, 1)
            this_x_range = x_range[i]
            target_label = dataset['target_label'][example_idx_list[i]]
            c = torch.zeros([1, self.num_outputs])
            c[0, label] = 1
            c[0, target_label] = -1
            new_c = [(c, self.rhs)]
            vnnlib.append([(this_x_range, new_c)])
        return vnnlib


class SpecificationRunnerup(Specification):
    def construct_vnnlib(self, dataset, X, x_range, example_idx_list):
        vnnlib = []
        for i in range(len(example_idx_list)):
            label = dataset['labels'][example_idx_list[i]].view(1, 1)
            this_x_range = x_range[i]
            runnerup = dataset['runnerup'][example_idx_list[i]]
            c = torch.zeros([1, self.num_outputs])
            c[0, label] = 1
            c[0, runnerup] = -1
            new_c = [(c, self.rhs)]
            vnnlib.append([(this_x_range, new_c)])
        return vnnlib


class SpecificationAllPositive(Specification):
    def construct_vnnlib(self, dataset, x_range, example_idx_list):
        vnnlib = []
        for i in range(len(example_idx_list)):
            this_x_range = x_range[i]
            c = torch.eye(self.num_outputs).unsqueeze(0)
            new_c = []
            for ii in range(self.num_outputs):
                new_c.append((c[:, ii], self.rhs))
            vnnlib.append([(this_x_range, new_c)])
        return vnnlib


def construct_vnnlib(dataset, example_idx_list):
    X = dataset['X']
    x_lower = x_upper = None
    if arguments.Config['specification']['type'] == 'lp':
        perturb_epsilon = dataset['eps']
        if type(perturb_epsilon) == list:
            # Each example has different perturbations.
            perturb_epsilon = torch.cat(perturb_epsilon)
            perturb_epsilon = perturb_epsilon[example_idx_list]
        assert perturb_epsilon is not None
        # FIXME why flatten?
        if arguments.Config['specification']['norm'] == float('inf'):
            if dataset.get('data_max', None) is None:
                # perturb_eps is already normalized.
                x_lower = (X[example_idx_list] - perturb_epsilon).flatten(1)
                x_upper = (X[example_idx_list] + perturb_epsilon).flatten(1)
            else:
                x_lower = (X[example_idx_list] - perturb_epsilon).clamp(
                    min=dataset['data_min']).flatten(1)
                x_upper = (X[example_idx_list] + perturb_epsilon).clamp(
                    max=dataset['data_max']).flatten(1)
            x_range = torch.stack([x_lower, x_upper], -1).numpy()
        else:
            # TODO create classes to handle it generally
            x_range = []
            for idx in example_idx_list:
                x_item = {
                    'X': X[idx],
                    'eps': dataset['eps'],
                    'norm': dataset['norm'],
                }
                if not isinstance(x_item['X'], torch.Tensor):
                    x_item['X'] = torch.tensor(x_item['X'])
                if 'eps_min' in dataset:
                    x_item['eps_min'] = dataset['eps_min']
                x_item['data_min'] = x_item['X'] - dataset['eps']
                x_item['data_max'] = x_item['X'] + dataset['eps']
                if dataset.get('data_min', None) is not None:
                    x_item['data_min'] = x_item['data_min'].clamp(
                        min=dataset['data_min'])
                if dataset.get('data_max', None) is not None:
                    x_item['data_max'] = x_item['data_max'].clamp(
                        min=dataset['data_max'])
                x_item['data_min'] = x_item['data_min']
                x_item['data_max'] = x_item['data_max']
                x_range.append(x_item)
    elif (arguments.Config['specification']['type'] == 'box' or
            # Some old config files use "bound"; keep for compatibility.
            arguments.Config['specification']['type'] == 'bound'):
        x_lower = dataset['data_min'].flatten(1)
        x_upper = dataset['data_max'].flatten(1)
        x_range = torch.stack([x_lower, x_upper], -1).numpy()
    else:
        raise ValueError('Unsupported perturbation type ' +
                         arguments.Config['specification']['type'])

    # TODO rename "robustness_type", since the verification objective may
    # not be related to robustness.
    robustness_type = arguments.Config['specification']['robustness_type']
    if robustness_type == 'verified-acc':
        specification = SpecificationVerifiedAcc()
    elif robustness_type == 'specify-target':
        specification = SpecificationTarget()
    elif robustness_type == 'runnerup':
        specification = SpecificationRunnerup()
    elif robustness_type == 'all-positive':
        specification = SpecificationAllPositive()
    else:
        raise ValueError(robustness_type)

    return specification.construct_vnnlib(
        dataset, x_range, example_idx_list)


def sort_targets_cls(batched_vnnlib, init_global_lb, init_global_ub, scores,
                     reference_alphas, lA, final_node_name, reverse=False):
    # TODO need minus rhs
    # To sort targets, this must be a classification task, and initial_max_domains
    # is set to 1.
    assert len(batched_vnnlib) == init_global_lb.shape[0] and init_global_lb.shape[1] == 1
    sorted_idx = scores.argsort(descending=reverse)
    batched_vnnlib = [batched_vnnlib[i] for i in sorted_idx]
    init_global_lb = init_global_lb[sorted_idx]
    init_global_ub = init_global_ub[sorted_idx]

    if reference_alphas is not None:
        for spec_dict in reference_alphas.values():
            for spec in spec_dict:
                if spec == final_node_name:
                    if spec_dict[spec].size()[1] > 1:
                        # correspond to multi-x case
                        spec_dict[spec] = spec_dict[spec][:, sorted_idx]
                    else:
                        spec_dict[spec] = spec_dict[spec][:, :, sorted_idx]

    if lA is not None:
        lA = {k: v[sorted_idx] for k, v in lA.items()}

    return batched_vnnlib, init_global_lb, init_global_ub, lA, sorted_idx


def trim_batch(model, init_global_lb, init_global_ub, reference_alphas_cp,
               orig_lower_bounds, orig_upper_bounds, reference_alphas, lA, property_idx,
               c, rhs):
    net = model.net
    optimize_disjuncts_separately = arguments.Config['solver']['optimize_disjuncts_separately']

    # FIXME (assigned to Kaidi, Jun 2023): this function might be wrong; it does not handles
    # the case with a few AND statements like yolo.
    # extract lower bound by (sorted) init_global_lb and batch size of initial_max_domains
    start_idx = property_idx * arguments.Config['bab']['initial_max_domains']
    lower_bounds = {}
    upper_bounds = {}
    if optimize_disjuncts_separately:
        for layer_name in orig_lower_bounds.keys():
            lower_bounds[layer_name] = orig_lower_bounds[layer_name][start_idx: start_idx + c.shape[0]]
            upper_bounds[layer_name] = orig_upper_bounds[layer_name][start_idx: start_idx + c.shape[0]]
        assert torch.all(lower_bounds[net.final_name] == init_global_lb[start_idx: start_idx + c.shape[0]])
        assert torch.all(upper_bounds[net.final_name] == init_global_ub[start_idx: start_idx + c.shape[0]])
    else:
        for layer_name in orig_lower_bounds.keys():
            lower_bounds[layer_name] = orig_lower_bounds[layer_name]
            upper_bounds[layer_name] = orig_upper_bounds[layer_name]
        lower_bounds[net.final_name] = init_global_lb[start_idx: start_idx + c.shape[0]]
        upper_bounds[net.final_name] = init_global_ub[start_idx: start_idx + c.shape[0]]
    if rhs.numel() > 1:
        if optimize_disjuncts_separately:
            raise NotImplementedError("Output constraints for disjunctions are not supported for rhs.numel() > 1")
        assert init_global_lb.numel() == rhs.numel()
        rhs = rhs[:, start_idx: start_idx + c.shape[0]]
    # trim reference slope by batch size of initial_max_domains accordingly
    if reference_alphas is not None:
        for m, spec_dict in reference_alphas.items():
            for spec in spec_dict:
                if spec == net.final_node().name:
                    if reference_alphas_cp[m][spec].size()[1] > 1:
                        # correspond to multi-x case
                        spec_dict[spec] = reference_alphas_cp[m][spec][
                            :, start_idx: start_idx + c.shape[0]]
                    else:
                        spec_dict[spec] = reference_alphas_cp[m][spec][
                            :, :, start_idx: start_idx + c.shape[0]]
    # trim lA by batch size of initial_max_domains accordingly
    if lA is not None:
        lA = {k: v[start_idx: start_idx + c.shape[0]] for k, v in lA.items()}
    return {
        'lA': lA, 'rhs': rhs, 'lower_bounds': lower_bounds, 'upper_bounds': upper_bounds
    }


def prune_by_idx(reference_alphas, init_verified_cond, final_name, lA_trim, x, data_min, data_max,
                 need_prune_lA, lower_bounds, upper_bounds, c):
    """
     Prune reference_alphas, lA_trim, x, data_min, data_max, lower_bounds, upper_bounds, c
     by init_verified_cond. Only keep unverified elements for next step bab or attack.
    """

    init_unverified_cond = ~init_verified_cond

    if reference_alphas is not None:
        LiRPANet.prune_reference_alphas(
            reference_alphas, init_unverified_cond, final_name)
    if need_prune_lA:
        lA_trim = LiRPANet.prune_lA(lA_trim, init_unverified_cond)

    if data_min.shape[0] > 1:
        # use [torch.where(~init_verified_cond)[0]] to prune x
        # when vnnlib has multiple different x
        # fixed: don't repeat x and then take [0:1]
        x, data_min, data_max = \
            x[torch.where(init_unverified_cond)[0]], \
            data_min[torch.where(init_unverified_cond)[0]], \
            data_max[torch.where(init_unverified_cond)[0]]

    lower_bounds[final_name] = lower_bounds[final_name][init_unverified_cond]
    upper_bounds[final_name] = upper_bounds[final_name][init_unverified_cond]
    c = c[torch.where(init_unverified_cond)[0]]

    return reference_alphas, lA_trim, x, data_min, data_max, lower_bounds, upper_bounds, c


def batch_vnnlib(vnnlib):
    """reorganize original vnnlib file, make x, c and rhs batch wise"""
    final_merged_rv = []

    init_d = {'x': [], 'c': [], 'rhs': [],
              'verify_criterion': [], 'attack_criterion': [] }
    target_labels = []

    for vnn in vnnlib:
        for mat, rhs in vnn[1]:
            if isinstance(vnn[0], dict):
                init_d['x'].append(vnn[0])
            else:
                init_d['x'].append(np.array(vnn[0]))
            init_d['c'].append(mat)
            init_d['rhs'].append(rhs)
            tmp_true_labels, tmp_target_labels = [], []
            for m in mat:

                target_label = np.where(m == -1)[-1]
                if len(target_label) != 0:
                    assert len(target_label) == 1
                    tmp_target_labels.append(target_label[0])
                else:
                    tmp_target_labels.append(None)

            target_labels.append(np.array(tmp_target_labels))

    if len(init_d['x']) == 0 or isinstance(init_d['x'][0], np.ndarray):
        # n, shape, 2; the batch dim n is necessary, even if n = 1
        init_d['x'] = np.array(init_d['x'])
    init_d['c'] = torch.concat(
        [(item if isinstance(item, torch.Tensor)
          else torch.tensor(item)).unsqueeze(0)
         for item in init_d['c']], dim=0)
    init_d['rhs'] = np.array(init_d['rhs'])  # n, n_output
    target_labels = np.array(target_labels)

    # batch_size = min(
    #         arguments.Config['solver']['batch_size'],
    #         arguments.Config['bab']['initial_max_domains'])
    # initial_max_domains can be much larger than batch_size if auto_enlarge_batch_size enabled
    batch_size = arguments.Config['bab']['initial_max_domains']

    total_batch = int(np.ceil(len(init_d['x']) / batch_size))
    print(f"Total VNNLIB file length: {len(init_d['x'])}, max property batch size: {batch_size}, total number of batches: {total_batch}")

    for i in range(total_batch):
        # [x, [(c, rhs, y, pidx)]], pidx can be none
        final_merged_rv.append([
            init_d['x'][i * batch_size: (i + 1) * batch_size],
            [(init_d['c'][i * batch_size: (i + 1) * batch_size],
              init_d['rhs'][i * batch_size: (i + 1) * batch_size],
              target_labels[i * batch_size: (i + 1) * batch_size]
            )]])

    return final_merged_rv


def sort_targets(batched_vnnlib, init_global_lb, init_global_ub,
                 attack_images=None, attack_margins=None, results=None,
                 model_incomplete=None):
    bab_attack_enabled = arguments.Config['bab']['attack']['enabled']
    sort_targets = arguments.Config['bab']['sort_targets']
    cplex_cuts = arguments.Config['bab']['cut']['enabled'] and arguments.Config['bab']['cut']['cplex_cuts']
    optimize_disjuncts_separately = arguments.Config['solver']['optimize_disjuncts_separately']
    reference_alphas = results.get('alpha', None)
    lA = results.get('lA', None)

    ret = None
    if bab_attack_enabled:
        # Sort specifications based on adversarial attack margins.
        ret = sort_targets_cls(
            batched_vnnlib, init_global_lb, init_global_ub, lA=lA,
            scores=attack_margins.flatten(), reference_alphas=reference_alphas,
            final_node_name=model_incomplete.net.final_node().name)
        attack_images = attack_images[:, :, ret[-1]]
    elif cplex_cuts:
        # need to sort pidx such that easier first according to initial alpha crown
        ret = sort_targets_cls(
            batched_vnnlib, init_global_lb, init_global_ub, lA=lA,
            scores=init_global_lb.flatten(), reference_alphas=reference_alphas,
            final_node_name=model_incomplete.net.final_node().name,
            reverse=True)
    elif sort_targets:
        # Sort specifications based on incomplete verifier bounds.
        ret = sort_targets_cls(
            batched_vnnlib, init_global_lb, init_global_ub, lA=lA,
            scores=init_global_lb.flatten(), reference_alphas=reference_alphas,
            final_node_name=model_incomplete.net.final_node().name)
    if ret:
        assert not optimize_disjuncts_separately, (
            "Sorting targets is currently not supported when disjuncts are optimized separately."
        )
        batched_vnnlib, init_global_lb, init_global_ub, lA = ret[:-1]

    return batched_vnnlib, init_global_lb, init_global_ub, lA, attack_images


def add_rhs_offset(vnnlib, rhs_offset):
    print('Add an offset to RHS for debugging:', rhs_offset)
    vnnlib = [
        (
            v[0],
            [(v[1][i][0], v[1][i][1] + rhs_offset)
            for i in range(len(v[1]))]
        )
        for v in vnnlib
    ]
    return vnnlib



import torch
import numpy as np

from CTRAIN.attacks import pgd_attack
from CTRAIN.train.certified.losses.ibp import get_ibp_loss

def get_mtl_ibp_loss(hardened_model, original_model, ptb, data, target, n_classes, criterion, alpha, return_bounds=False, return_stats=False, restarts=1, step_size=.2, n_steps=200, pgd_ptb=None, early_stopping=False, decay_checkpoints=(), decay_factor=.1, device='cuda'):    
    hardened_model.eval()
    original_model.eval()
    x_adv = pgd_attack(
        model=hardened_model,
        data=data,
        target=target,
        x_L=pgd_ptb.x_L,
        x_U=pgd_ptb.x_U,
        restarts=restarts,
        step_size=step_size,
        n_steps=n_steps,
        early_stopping=early_stopping,
        device=device,
        decay_factor=decay_factor,
        decay_checkpoints=decay_checkpoints
    )
    
    hardened_model.train()
    original_model.train()
    
    adv_output = hardened_model(x_adv)
    adv_loss = criterion(adv_output, target).mean()

    robust_loss, lb, ub = get_ibp_loss(
        hardened_model=hardened_model,
        ptb=ptb,
        data=data,
        target=target,
        n_classes=n_classes,
        criterion=criterion,
        return_bounds=True
    )
    
    loss = alpha * robust_loss + (1 - alpha) * adv_loss
    
    return_tuple = (loss,)
    
    if return_bounds:
        assert False, "Return bounds is not implemented for MTL-IBP"
    elif return_stats:
        robust_err = torch.sum((lb < 0).any(dim=1)).item() / data.size(0)
        adv_err = torch.sum(torch.argmax(adv_output, dim=1) != target).item() / data.size(0)
        return_tuple = return_tuple + (robust_err, adv_err)
    
    return return_tuple


# def get_cc_ibp_loss(hardened_model, ptb, data, target, n_classes, criterion, alpha, return_bounds=False, restarts=1, step_size=.2, n_steps=200, early_stopping=False):
#     # TODO: use different (higher) eps for PGD
    
#     x_adv = pgd_attack(
#         model=hardened_model,
#         data=data,
#         target=target,
#         x_L=ptb.x_L,
#         x_U=ptb.x_U,
#         restarts=restarts,
#         step_size=step_size,
#         n_steps=n_steps,
#         early_stopping=early_stopping
#     )
    
#     # Pass x_adv through model s.t. IBP uses batch norm statistics from x_adv
#     adv_pred = hardened_model(x_adv)
#     adv_lb = get
    
        
#     ibp_lb, ibp_up = bound_ibp(
#         model=hardened_model,
#         ptb=ptb,
#         data=data,
#         target=target,
#         n_classes=n_classes,
#     )

    
#     if return_bounds:
#         assert False, "Return bounds is not implemented for MTL-IBP"
#     else:
#         return (1-alpha) * adv_loss + alpha * certified_loss


#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import os
import collections
import gzip
from ast import literal_eval
import torch
import numpy as np

import onnx2pytorch
import onnx
import onnxruntime as ort
import onnxoptimizer
from skl2onnx.helpers.onnx_helper import select_model_inputs_outputs
from skl2onnx.helpers.onnx_helper import save_onnx_model
from skl2onnx.helpers.onnx_helper import enumerate_model_node_outputs
from onnx_opt import compress_onnx

import warnings
import importlib
from functools import partial

import arguments

from model_defs import *
from utils import expand_path


def Customized(def_file, callable_name, *args, **kwargs):
    """Fully customized model or dataloader."""
    if def_file.endswith('.py'):
        # Use relatively path w.r.t. to the configuration file
        if arguments.Config['general']['root_path']:
            path = os.path.join(
                expand_path(arguments.Config['general']['root_path']), def_file)
        elif arguments.Config.file:
            path = os.path.join(os.path.dirname(arguments.Config.file), def_file)
        else:
            path = def_file
        spec = importlib.util.spec_from_file_location('customized', path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    else:
        try:
            # Customized loaders should be in "custom" folder.
            module = importlib.import_module(f'custom.{def_file}')
        except ModuleNotFoundError:
            # If not found, try the current folder.
            module = importlib.import_module(f'{def_file}')
            warnings.warn(  # Old config files may refer to custom loaders in the root folder.
                    f'Customized loaders "{def_file}" should be inside the "custom" folder.')
    # Load model from a specified file.
    model_func = getattr(module, callable_name)
    customized_func = partial(model_func, *args, **kwargs)
    # We need to return a Callable which returns the model.
    return customized_func


def deep_update(d, u):
    """Update a dictionary based another dictionary, recursively.

    (https://stackoverflow.com/a/3233356).
    """
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def unzip_and_optimize_onnx(path, onnx_optimization_flags='none'):
    if onnx_optimization_flags == 'none':
        if path.endswith('.gz'):
            onnx_model = onnx.load(gzip.GzipFile(path))
        else:
            onnx_model = onnx.load(path)
        return onnx_model
    else:
        print(f'Onnx optimization with flag: {onnx_optimization_flags}')
        npath = path + '.optimized'
        if os.path.exists(npath):
            print(f'Found existed optimized onnx model at {npath}')
            return onnx.load(npath)
        else:
            print(f'Generate optimized onnx model to {npath}')
            if path.endswith('.gz'):
                onnx_model = onnx.load(gzip.GzipFile(path))
            else:
                onnx_model = onnx.load(path)
            return compress_onnx(onnx_model, path, npath, onnx_optimization_flags, debug=True)


def inference_onnx(path, input):
    # Workaround for onnx bug, see issue #150
    options = ort.SessionOptions()
    options.inter_op_num_threads = 1
    options.intra_op_num_threads = 1
    sess = ort.InferenceSession(unzip_and_optimize_onnx(path).SerializeToString(),
                                sess_options=options)
    assert len(sess.get_inputs()) == len(sess.get_outputs()) == 1
    res = sess.run(None, {sess.get_inputs()[0].name: input})[0]
    return res


@torch.no_grad()
def load_model_onnx(path, quirks=None, x=None):
    onnx_optimization_flags = arguments.Config['model']['onnx_optimization_flags']
    if arguments.Config['model']['cache_onnx_conversion']:
        path_cache = f'{path}.cache'
        if os.path.exists(path_cache):
            print(f'Loading converted model from {path_cache}')
            return torch.load(path_cache)
    quirks = {} if quirks is None else quirks
    if arguments.Config['model']['onnx_quirks']:
        try:
            config_quirks = literal_eval(arguments.Config['model']['onnx_quirks'])
        except ValueError:
            print('ERROR: onnx_quirks '
                  f'{arguments.Config["model"]["onnx_quirks"]}'
                  'cannot be parsed!')
            raise
        assert isinstance(config_quirks, dict)
        deep_update(quirks, config_quirks)
    print(f'Loading onnx {path} with quirks {quirks}')

    onnx_model = unzip_and_optimize_onnx(path, onnx_optimization_flags)

    if arguments.Config["model"]["input_shape"] is None:
        # find the input shape from onnx_model generally
        # https://github.com/onnx/onnx/issues/2657
        input_all = [node.name for node in onnx_model.graph.input]
        input_initializer = [node.name for node in onnx_model.graph.initializer]
        net_feed_input = list(set(input_all) - set(input_initializer))
        net_feed_input = [node for node in onnx_model.graph.input if node.name in net_feed_input]

        if len(net_feed_input) != 1:
            # in some rare case, we use the following way to find input shape
            # but this is not always true (collins-rul-cnn)
            net_feed_input = [onnx_model.graph.input[0]]

        onnx_input_dims = net_feed_input[0].type.tensor_type.shape.dim
        onnx_shape = tuple(d.dim_value for d in onnx_input_dims)
    else:
        # User specify input_shape
        onnx_shape = arguments.Config['model']['input_shape']

    # remove batch information
    # for nn4sys pensieve parallel, the first dimension of the input size is not batch, do not remove
    if onnx_shape[0] <= 1:
        onnx_shape = onnx_shape[1:]

    try:
        pytorch_model = onnx2pytorch.ConvertModel(
            onnx_model, experimental=True, quirks=quirks)
    except TypeError as e:
        print('\n\nA possible onnx2pytorch version error!')
        print('If you see "unexpected keyword argument \'quirks\'", that indicates your onnx2pytorch version is incompatible.')
        print('Please uninstall onnx2pytorch in your python environment (e.g., run "pip uninstall onnx2pytorch"), and then reinstall using:\n')
        print('pip install git+https://github.com/KaidiXu/onnx2pytorch@fe7281b9b6c8c28f61e72b8f3b0e3181067c7399\n\n')
        print('The error below may not a bug of alpha-beta-CROWN. See instructions above.')
        raise(e)
    pytorch_model.eval()
    pytorch_model.to(dtype=torch.get_default_dtype())

    conversion_check_result = True
    try:
        # check conversion correctness
        # FIXME dtype of dummy may not match the onnx model, which can cause runtime error
        if x is not None:
            x = x.reshape(1, *onnx_shape)
        else:
            x = torch.randn([1, *onnx_shape])
        output_pytorch = pytorch_model(x).numpy()
        try:
            if arguments.Config['model']['check_optmized']:
                output_onnx = inference_onnx(path+'.optimized', x.numpy())
            else:
                output_onnx = inference_onnx(path, x.numpy())
        except ort.capi.onnxruntime_pybind11_state.InvalidArgument:
            # ONNX model might have shape problems. Remove the batch dimension and try again.
            output_onnx = inference_onnx(path, x.numpy().squeeze(0))
        if 'remove_relu_in_last_layer' in onnx_optimization_flags:
            output_pytorch = output_pytorch.clip(min=0)
        conversion_check_result = np.allclose(
            output_pytorch, output_onnx, 1e-4, 1e-5)
    except:  # pylint: disable=broad-except
        warnings.warn('Not able to check model\'s conversion correctness')
        print('\n*************Error traceback*************')
        import traceback; print(traceback.format_exc())
        print('*****************************************\n')
    if not conversion_check_result:
        print('\n**************************')
        print('Model might not be converted correctly. Please check onnx conversion carefully.')
        print('Output by pytorch:', output_pytorch)
        print('Output by onnx:', output_onnx)
        print('Max error:', torch.tensor(output_pytorch - output_onnx).abs().max())
        print('**************************\n')
        if arguments.Config["model"]["debug_onnx"]:
            debug_onnx(onnx_model, pytorch_model, x.numpy())

    if arguments.Config["model"]["cache_onnx_conversion"]:
        torch.save((pytorch_model, onnx_shape), path_cache)

    # TODO merge into the unzip_and_optimize_onnx()
    if arguments.Config["model"]["flatten_final_output"]:
        pytorch_model = nn.Sequential(pytorch_model, nn.Flatten())

    return pytorch_model, onnx_shape


def debug_onnx(onnx_model, pytorch_model, dummy_input):
    path_tmp = '/tmp/debug.onnx'

    output_onnx = {}
    for node in enumerate_model_node_outputs(onnx_model):
        print('Inferencing onnx node:', node)
        save_onnx_model(select_model_inputs_outputs(onnx_model, node), path_tmp)
        optimized_model = onnxoptimizer.optimize(
            onnx.load(path_tmp),
            ["extract_constant_to_initializer",
             "eliminate_unused_initializer"])
        sess = ort.InferenceSession(optimized_model.SerializeToString())
        output_onnx[node] = torch.tensor(sess.run(
            None, {sess.get_inputs()[0].name: dummy_input})[0])

    print('Inferencing the pytorch model')
    output_pytorch = pytorch_model(
        torch.tensor(dummy_input), return_all_nodes=True)

    for k in output_pytorch:
        if k == sess.get_inputs()[0].name:
            continue
        print(k, output_onnx[k].shape)
        close = torch.allclose(output_onnx[k], output_pytorch[k])
        print('  close?', close)
        if not close:
            print('  max error', (output_onnx[k] - output_pytorch[k]).abs().max())

    import pdb; pdb.set_trace()


def load_model(weights_loaded=True):
    """
    Load the model architectures and weights
    """

    assert arguments.Config["model"]["name"] is None or arguments.Config["model"]["onnx_path"] is None, (
        "Conflict detected! User should specify model path by either --model or --onnx_path! "
        "The cannot be both specified.")

    assert arguments.Config["model"]["name"] is not None or arguments.Config["model"]["onnx_path"] is not None, (
        "No model is loaded, please set --model <modelname> for pytorch model or --onnx_path <filename> for onnx model.")

    if arguments.Config['model']['name'] is not None:
        # You can customize this function to load your own model based on model name.
        try:
            model_ori = eval(arguments.Config['model']['name'])()  # pylint: disable=eval-used
        except Exception:  # pylint: disable=broad-except
            print(f'Cannot load pytorch model definition "{arguments.Config["model"]["name"]}()". '
                  f'"{arguments.Config["model"]["name"]}()" must be a callable that returns a torch.nn.Module object.')
            import traceback
            traceback.print_exc()
            exit()
        model_ori.eval()

        if not weights_loaded:
            return model_ori

        if arguments.Config["model"]["path"] is not None:
            # Load pytorch model
            # You can customize this function to load your own model based on model name.
            sd = torch.load(expand_path(arguments.Config["model"]["path"]),
                            map_location=torch.device('cpu'))
            if 'state_dict' in sd:
                sd = sd['state_dict']
            if isinstance(sd, list):
                sd = sd[0]
            if not isinstance(sd, dict):
                raise NotImplementedError("Unknown model format, please modify model loader yourself.")
            try:
                model_ori.load_state_dict(sd)
            except RuntimeError:
                print('Failed to load the model')
                print('Keys in the state_dict of model_ori:')
                print(list(model_ori.state_dict().keys()))
                print('Keys in the state_dict trying to load:')
                print(list(sd.keys()))
                raise

    elif arguments.Config["model"]["onnx_path"] is not None:
        # Load onnx model
        model_ori, _ = load_model_onnx(expand_path(
            arguments.Config["model"]["onnx_path"]))

    else:
        print("Warning: pretrained model path is not given!")

    print(model_ori)
    print('Parameters:')
    for p in model_ori.named_parameters():
        print(f'  {p[0]}: shape {p[1].shape}')

    return model_ori



import copy
import types
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

from auto_LiRPA.bound_general import BoundedModule

from auto_LiRPA.operators.normalization import BoundBatchNormalization

from auto_LiRPA.operators.solver_utils import grb

from auto_LiRPA.linear_bound import LinearBound

from auto_LiRPA.operators.constant import BoundConstant

from auto_LiRPA.operators.leaf import BoundParams

from auto_LiRPA.patches import Patches, inplace_unfold

from auto_LiRPA.operators.base import Interval


def split_network(model, block_sizes, network_input, device):
    # TODO: Add assertions for robustness
    start = 0
    original_blocks = []
    network_input = network_input.to(device)
    for size in block_sizes:
        end = start + size
        abs_block = nn.Sequential(model.layers[start:end])
        original_blocks.append(abs_block)
        
        output_shape = abs_block(network_input).shape
        network_input = torch.zeros(output_shape).to(device)
        
        start = end
    return original_blocks

def freeze_batch_norm(model):
    def _check_unused_mean_or_var(self):
        if not hasattr(self, 'frozen') or not self.frozen:
            # Check if either mean or var is opted out
            if not self.use_mean:
                self.current_mean = torch.zeros_like(self.current_mean)
            if not self.use_var:
                self.current_var = torch.ones_like(self.current_var)
        else:
            pass

    def forward(self, x, w, b, m, v):
        if len(x.shape) == 2:
            self.patches_start = False
        if self.training and not self.frozen:
            dim = [0] + list(range(2, x.ndim))
            self.current_mean = x.mean(dim)
            self.current_var = x.var(dim, unbiased=False)
        else:
            self.current_mean = m.data
            self.current_var = v.data
        self._check_unused_mean_or_var()
        if not self.use_affine:
            w = torch.ones_like(w)
            b = torch.zeros_like(b)
        result = F.batch_norm(x, m, v, w, b, self.training, self.momentum, self.eps)
        if not self.use_mean or not self.use_var:
            # If mean or variance is disabled, recompute the output from self.current_mean
            # and self.current_var instead of using standard F.batch_norm.
            w = w / torch.sqrt(self.current_var + self.eps)
            b = b - self.current_mean * w
            shape = (1, -1) + (1,) * (x.ndim - 2)
            result = w.view(*shape) * x + b.view(*shape)
        return result
    
    for module in model.modules():
        if isinstance(module, BoundBatchNormalization):
            assert not hasattr(module, 'orig_momentum'), "Batch Norm is already frozen! Please run unfreeze_batch_norm before trying to freeze it!"
            module.orig_momentum = module.momentum
            module.momentum = 0.0
            module._check_unused_mean_or_var = types.MethodType(_check_unused_mean_or_var, module)
            module.forward = types.MethodType(forward, module)
            module.frozen = True
            module.use_mean = False
            module.use_var = False

def unfreeze_batch_norm(model):
    for module in model.modules():
        if isinstance(module, BoundBatchNormalization):
            assert hasattr(module, 'orig_momentum'), "Please run freeze_batch_norm before trying to unfreeze it"
            module.momentum = module.orig_momentum
            del module.orig_momentum
            module.use_mean = True
            module.use_var = True
            module.frozen = False


def reset_batch_norms_to_data_statistics(model, dataloader, device='cuda'):
    '''
    Use population statistics to reset the BN layers in the model.
    '''
    model.to(device)
    bn_list = [m for m in model.modules() if isinstance(m, BoundBatchNormalization)]
    if len(bn_list) == 0:
        return model
    model.train()
    momentum_list = [m.momentum for m in bn_list]
    num_batches_tracked = 0
    for x, _ in dataloader:
        x = x.to(device)
        num_batches_tracked += 1
        with torch.no_grad():
            for m in bn_list:
                m.momentum = 1 / num_batches_tracked
            model(x)
    for m in bn_list:
        m.momentum = momentum_list.pop(0)



from ConfigSpace import Categorical, ConfigurationSpace, EqualsCondition, Float, Integer, Constant
import math

import torch


def get_config_space(model_wrapper, epochs, eps, defaults=dict()):
    if model_wrapper.cert_train_method == 'shi':
        return build_shi_config_space(epochs, eps, defaults=defaults)
    elif model_wrapper.cert_train_method == 'crown_ibp':
        return build_crown_ibp_config_space(epochs, eps, defaults=defaults)
    elif model_wrapper.cert_train_method == 'sabr':
        return build_sabr_config_space(epochs, eps, defaults=defaults)
    elif model_wrapper.cert_train_method == 'taps':
        return build_taps_config_space(epochs, eps, defaults=defaults)
    elif model_wrapper.cert_train_method == 'staps':
        return build_staps_config_space(epochs, eps, defaults=defaults)
    elif model_wrapper.cert_train_method == 'mtl_ibp':
        return build_mtl_ibp_config_space(epochs, eps, defaults=defaults)
    else:
        assert False, 'Unknown model wrapper!'

def get_base_config(epochs, eps, defaults=dict()):
    # TODO: maybe make it possible to pass lr decay epochs as tuple as done in wrapper
    # TODO: maybe convert default torch.optim instance to string

    return ConfigurationSpace(
        name='base',
        space={
            'warm_up_epochs': Integer('warm_up_epochs', (0, 5), default=defaults.get('warm_up_epochs')),
            'ramp_up_epochs': Integer('ramp_up_epochs', (10, math.ceil(0.75*epochs)), default=defaults.get('ramp_up_epochs')),
            'lr_decay_factor': Float('lr_decay_factor', (1e-5, 0.9), log=True, default=defaults.get('lr_decay_factor')),
            'lr_decay_epoch_1': Integer('lr_decay_epoch_1', (0, math.ceil(.5 * epochs)), default=defaults.get('lr_decay_epoch_1')),
            'lr_decay_epoch_2': Integer('lr_decay_epoch_2', (0, math.ceil(.25 * epochs)), default=defaults.get('lr_decay_epoch_2')),
            'l1_reg_weight': Float('l1_reg_weight', (1e-8, 1e-4), log=True, default=defaults.get('l1_reg_weight')),
            'shi_reg_weight': Float("shi_reg_weight", (0.0, 1.0), default=defaults.get('shi_reg_weight')),
            'shi_reg_decay': Categorical('shi_reg_decay', [True, False], default=defaults.get('shi_reg_decay')),
            'train_eps_factor': Float("train_eps_factor", (1, 2), default=defaults.get('train_eps_factor')),
            'optimizer_func': Categorical('optimizer_func', ['adam', 'adamw', 'radam'], default=defaults.get('optimizer_func')),
            'learning_rate': Float('learning_rate', (1e-5, 0.1), log=True, default=defaults.get('lr'))
        }
    )

def build_shi_config_space(epochs, eps, include_base_config=True, defaults=dict()):
    if include_base_config:
        config_space = get_base_config(epochs, eps, defaults=defaults)
    else:
        config_space = ConfigurationSpace()
    
    shi_config_space = ConfigurationSpace(
        name='shi',
        space={
            'start_kappa': Float('start_kappa', (0.5, 1), default=defaults.get('start_kappa')),
            'end_kappa': Float('end_kappa', (0, 1), default=defaults.get('end_kappa')) # we multiply start kappa by this factor to obtain the real end_kappa
        }
    )
    config_space.add_configuration_space('shi', shi_config_space)

    return config_space 


def build_crown_ibp_config_space(epochs, eps, include_base_config=True, defaults=dict()):
    if include_base_config:
        config_space = get_base_config(epochs, eps, defaults=defaults)
    else:
        config_space = ConfigurationSpace()

    crown_ibp_config_space = ConfigurationSpace(
        name='crown_ibp',
        space={
            'start_kappa': Float('start_kappa', (0.5, 1), default=defaults.get('start_kappa')),
            'end_kappa': Float('end_kappa', (0, 1), default=defaults.get('end_kappa')), # we multiply start kappa by this factor to obtain the real end_kappa
            'start_beta': Constant('start_beta', 1.0),
            'end_beta': Float('end_beta', (0, 1), default=defaults.get('end_beta'))
        }
    )
    config_space.add_configuration_space('crown_ibp', crown_ibp_config_space)

    return config_space 


def build_sabr_config_space(epochs, eps, include_base_config=True, defaults=dict()):
    if include_base_config:
        config_space = get_base_config(epochs, eps, defaults=defaults)
    else:
        config_space = ConfigurationSpace()

    sabr_config_space = ConfigurationSpace(
        name='sabr',
        space={
            'subselection_ratio': Float('subselection_ratio', (0.01, 0.9), default=defaults.get('subselection_ratio')),
            'pgd_steps': Integer('pgd_steps', (1, 10), default=defaults.get('pgd_steps')),
            'pgd_alpha': Float('pgd_alpha', (0.1, 2), default=defaults.get('pgd_alpha')),
            'pgd_restarts': Constant('pgd_restarts', 1), # fixed number since it can get very expensive with more restarts
            # we do not optimise the pgd decay
        },

    )
    config_space.add_configuration_space('sabr', sabr_config_space)

    return config_space 

def build_taps_config_space(epochs, eps, include_base_config=True, defaults=dict()):
    if include_base_config:
        config_space = get_base_config(epochs, eps, defaults=defaults)
    else:
        config_space = ConfigurationSpace()
    
    taps_config_space = ConfigurationSpace(
        name='taps',
        space={
            'pgd_steps': Integer('pgd_steps', (1, 10), default=defaults.get('pgd_steps')),
            'pgd_alpha': Float('pgd_alpha', (0.1, 2), default=defaults.get('pgd_alpha')),
            'pgd_restarts': Constant('pgd_restarts', 1), # fixed number since it can get very expensive with more restarts
            'gradient_expansion_alpha': Float('gradient_expansion_alpha', (1, 10), default=defaults.get('gradient_expansion_alpha')),
            'block_split_point': Float('block_split_point', (0.01, 0.99), default=defaults.get('block_split_point'))
            # we do not optimise the pgd decay
        },

    )
    config_space.add_configuration_space('taps', taps_config_space)

    return config_space 

def build_staps_config_space(epochs, eps, include_base_config=True, defaults=dict()):
    config_space = build_taps_config_space(epochs, eps, defaults=defaults, include_base_config=include_base_config)

    sabr_config_space = ConfigurationSpace(
        name='sabr',
        space={
            'subselection_ratio': Float('subselection_ratio', (0.01, 0.9), default=defaults.get('subselection_ratio')),
            'pgd_steps': Integer('pgd_steps', (1, 10), default=defaults.get('pgd_steps')),
            'pgd_alpha': Float('pgd_alpha', (0.1, 2), default=defaults.get('pgd_alpha')),
            'pgd_restarts': Constant('pgd_restarts', 1), # fixed number since it can get very expensive with more restarts
        },

    )
    config_space.add_configuration_space('sabr', sabr_config_space)

    return config_space 


def build_mtl_ibp_config_space(epochs, eps, include_base_config=True, defaults=dict()):
    if include_base_config:
        config_space = get_base_config(epochs, eps, defaults=defaults)
    else:
        config_space = ConfigurationSpace()

    mtl_ibp_config_space = ConfigurationSpace(
        name='mtl_ibp',
        space={
            'mtl_ibp_alpha': Float('mtl_ibp_alpha', (0.001, 0.7), log=True, default=defaults.get('mtl_ibp_alpha')),
            'pgd_steps': Integer('pgd_steps', (1, 10), default=defaults.get('pgd_steps')),
            'pgd_alpha': Float('pgd_alpha', (0.1, 2), default=defaults.get('pgd_alpha')),
            'pgd_restarts': Constant('pgd_restarts', 1), # fixed number since it can get very expensive with more restarts
            'mtl_ibp_eps_factor': Float("mtl_ibp_eps_factor", (1, 3), default=defaults.get('mtl_ibp_eps_factor')),
        },

    )
    config_space.add_configuration_space('mtl_ibp', mtl_ibp_config_space)

    return config_space 

def get_combined_config_space(epoch, eps, defaults=dict(), included_methods=['shi', 'crown_ibp', 'sabr', 'mtl_ibp', 'taps', 'staps']):
    base_config_space = get_base_config(epoch, eps, defaults=defaults)
    base_config_space.add_hyperparameter(
        Categorical('cert_train_method', included_methods)
    )
    if 'shi' in included_methods:
        shi_config = build_shi_config_space(epoch, eps, include_base_config=False, defaults=defaults)
        base_config_space.add_configuration_space('shi', shi_config)
        base_config_space.add_conditions(
            [EqualsCondition(base_config_space[param], base_config_space['cert_train_method'], 'shi') for param in base_config_space if 'shi:' in param],
        )
    if 'crown_ibp' in included_methods:
        crown_ibp_config = build_crown_ibp_config_space(epoch, eps, include_base_config=False, defaults=defaults)
        base_config_space.add_configuration_space('crown_ibp', crown_ibp_config)
        base_config_space.add_conditions(
            [EqualsCondition(base_config_space[param], base_config_space['cert_train_method'], 'crown_ibp') for param in base_config_space if 'crown_ibp:' in param],
        )
    if 'sabr' in included_methods:
        sabr_config = build_sabr_config_space(epoch, eps, include_base_config=False, defaults=defaults)
        base_config_space.add_configuration_space('sabr', sabr_config)
        base_config_space.add_conditions(
            [EqualsCondition(base_config_space[param], base_config_space['cert_train_method'], 'sabr') for param in base_config_space if 'sabr:' in param],
        )
    if 'taps' in included_methods:
        taps_config = build_taps_config_space(epoch, eps, include_base_config=False, defaults=defaults)
        base_config_space.add_configuration_space('taps', taps_config)
        base_config_space.add_conditions(
            [EqualsCondition(base_config_space[param], base_config_space['cert_train_method'], 'taps') for param in base_config_space if 'taps:' in param],
        )
    if 'staps' in included_methods:
        staps_config = build_staps_config_space(epoch, eps, include_base_config=False, defaults=defaults)
        base_config_space.add_configuration_space('staps', staps_config)
        base_config_space.add_conditions(
            [EqualsCondition(base_config_space[param], base_config_space['cert_train_method'], 'staps') for param in base_config_space if 'staps:' in param],
        )
    if 'mtl_ibp' in included_methods:
        mtl_ibp_config = build_mtl_ibp_config_space(epoch, eps, include_base_config=False, defaults=defaults)
        base_config_space.add_configuration_space('mtl_ibp', mtl_ibp_config)
        base_config_space.add_conditions(
            [EqualsCondition(base_config_space[param], base_config_space['cert_train_method'], 'mtl_ibp') for param in base_config_space if 'mtl_ibp:' in param],
        )

    return base_config_space
    
    
    




#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
"""
This file shows how to use customized models and customized dataloaders.

Use the example configuration:
python abcrown.py --config exp_configs/tutorial_examples/custom_model_data_example.yaml
"""

import os
import torch
from torch import nn
from torchvision import transforms
from torchvision import datasets
import arguments


def simple_conv_model(in_channel, out_dim):
    """Simple Convolutional model."""
    model = nn.Sequential(
        nn.Conv2d(in_channel, 16, 4, stride=2, padding=0),
        nn.ReLU(),
        nn.Conv2d(16, 32, 4, stride=2, padding=0),
        nn.ReLU(),
        nn.Flatten(),
        nn.Linear(32*6*6,100),
        nn.ReLU(),
        nn.Linear(100, out_dim)
    )
    return model


def two_relu_toy_model(in_dim=2, out_dim=2):
    """A very simple model, 2 inputs, 2 ReLUs, 2 outputs"""
    model = nn.Sequential(
        nn.Linear(in_dim, 2),
        nn.ReLU(),
        nn.Linear(2, out_dim)
    )
    # [relu(x+2y)-relu(2x+y)+2, 0*relu(2x-y)+0*relu(-x+y)]
    model[0].weight.data = torch.tensor([[1., 2.], [2., 1.]])
    model[0].bias.data = torch.tensor([0., 0.])
    model[2].weight.data = torch.tensor([[1., -1.], [0., 0.]])
    model[2].bias.data = torch.tensor([2., 0.])
    return model


def simple_box_data(spec):
    """a customized box data: x=[-1, 1], y=[-1, 1]"""
    eps = spec["epsilon"]
    if eps is None:
        eps = 2.
    X = torch.tensor([[0., 0.]]).float()
    labels = torch.tensor([0]).long()
    eps_temp = torch.tensor(eps).reshape(1, -1)
    data_max = torch.tensor(10.).reshape(1, -1)
    data_min = torch.tensor(-10.).reshape(1, -1)
    return X, labels, data_max, data_min, eps_temp


def box_data(dim, low=0., high=1., segments=10, num_classes=10, eps=None):
    """Generate fake datapoints."""
    step = (high - low) / segments
    data_min = torch.linspace(low, high - step, segments).unsqueeze(1).expand(segments, dim)  # Per element lower bounds.
    data_max = torch.linspace(low + step, high, segments).unsqueeze(1).expand(segments, dim)  # Per element upper bounds.
    X = (data_min + data_max) / 2.  # Fake data.
    labels = torch.remainder(torch.arange(0, segments, dtype=torch.int64), num_classes)  # Fake label.
    eps = None  # Lp norm perturbation epsilon. Not used, since we will return per-element min and max.
    return X, labels, data_max, data_min, eps


def cifar10(spec, use_bounds=False):
    """Example dataloader. For MNIST and CIFAR you can actually use existing ones in utils.py."""
    eps = spec["epsilon"]
    assert eps is not None
    database_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'datasets')
    # You can access the mean and std stored in config file.
    mean = torch.tensor(arguments.Config["data"]["mean"])
    std = torch.tensor(arguments.Config["data"]["std"])
    normalize = transforms.Normalize(mean=mean, std=std)
    test_data = datasets.CIFAR10(database_path, train=False, download=True, transform=transforms.Compose([transforms.ToTensor(), normalize]))
    # Load entire dataset.
    testloader = torch.utils.data.DataLoader(test_data, batch_size=10000, shuffle=False, num_workers=4)
    X, labels = next(iter(testloader))
    if use_bounds:
        # Option 1: for each example, we return its element-wise lower and upper bounds.
        # If you use this option, set --spec_type ("specifications"->"type" in config) to 'bound'.
        absolute_max = torch.reshape((1. - mean) / std, (1, -1, 1, 1))
        absolute_min = torch.reshape((0. - mean) / std, (1, -1, 1, 1))
        # Be careful with normalization.
        new_eps = torch.reshape(eps / std, (1, -1, 1, 1))
        data_max = torch.min(X + new_eps, absolute_max)
        data_min = torch.max(X - new_eps, absolute_min)
        # In this case, the epsilon does not matter here.
        ret_eps = None
    else:
        # Option 2: return a single epsilon for all data examples, as well as clipping lower and upper bounds.
        # Set data_max and data_min to be None if no clip. For CIFAR-10 we clip to [0,1].
        data_max = torch.reshape((1. - mean) / std, (1, -1, 1, 1))
        data_min = torch.reshape((0. - mean) / std, (1, -1, 1, 1))
        if eps is None:
            raise ValueError('You must specify an epsilon')
        # Rescale epsilon.
        ret_eps = torch.reshape(eps / std, (1, -1, 1, 1))
    return X, labels, data_max, data_min, ret_eps


def simple_cifar10(spec):
    """Example dataloader. For MNIST and CIFAR you can actually use existing ones in utils.py."""
    eps = spec["epsilon"]
    assert eps is not None
    database_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'datasets')
    # You can access the mean and std stored in config file.
    mean = torch.tensor(arguments.Config["data"]["mean"])
    std = torch.tensor(arguments.Config["data"]["std"])
    normalize = transforms.Normalize(mean=mean, std=std)
    test_data = datasets.CIFAR10(database_path, train=False, download=True,\
            transform=transforms.Compose([transforms.ToTensor(), normalize]))
    # Load entire dataset.
    testloader = torch.utils.data.DataLoader(test_data,\
            batch_size=10000, shuffle=False, num_workers=4)
    X, labels = next(iter(testloader))
    # Set data_max and data_min to be None if no clip. For CIFAR-10 we clip to [0,1].
    data_max = torch.reshape((1. - mean) / std, (1, -1, 1, 1))
    data_min = torch.reshape((0. - mean) / std, (1, -1, 1, 1))
    if eps is None:
        raise ValueError('You must specify an epsilon')
    # Rescale epsilon.
    ret_eps = torch.reshape(eps / std, (1, -1, 1, 1))
    return X, labels, data_max, data_min, ret_eps



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import torch
import arguments

def customized_gtrsb_adv_example_finalizer(model_ori, x, best_deltas, data_max, data_min, C_mat, rhs_mat, cond_mat):
    # x and best_deltas has shape (batch, c, h, w).
    # data_min and data_max have shape (batch, spec, c, h, w).
    attack_image = torch.max(torch.min((x + best_deltas).unsqueeze(1), data_max), data_min)
    assert (attack_image >= data_min).all()
    assert (attack_image <= data_max).all()

    attack_output = model_ori(attack_image[:, 0, :]).repeat(1, attack_image.shape[1], 1)
    # [batch_size, num_or_spec, out_dim]

    if arguments.Config['general']['save_output']:
        arguments.Globals['out']['pred_adv'] = attack_output[0][0].cpu()

    # only print out the first two random start outputs of the first two examples.
    print("Adv example prediction (first 2 examples and 2 restarts):\n", attack_output[:2,:2])

    attack_output_repeat = attack_output.unsqueeze(1).repeat_interleave(torch.tensor(cond_mat[0], device=x.device), dim=2)
    # [num_example, num_restarts, num_spec, num_output]

    C_mat = C_mat.view(C_mat.shape[0], 1, -1, C_mat.shape[-1])
    # [num_example, 1, num_spec, num_output]
    rhs_mat = rhs_mat.view(rhs_mat.shape[0], 1, -1)
    # [num_example, 1, num_spec]

    attack_margin = (C_mat * attack_output_repeat).sum(-1) - rhs_mat
    # [num_example, num_restarts, num_spec]

    if arguments.Config['general']['save_output']:
        arguments.Globals['out']['attack_margin'] = attack_margin.cpu()

    print("PGD attack margin (first 2 examples and 10 specs):\n", attack_margin[:2, :, :10])
    print("number of violation: ", (attack_margin < 0).sum().item())
    # print the first 10 specifications for the first 2 examples

    return attack_image, attack_output, attack_margin



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import torch
import torch.nn as nn
import numpy as np
from auto_LiRPA import BoundedTensor
from auto_LiRPA.bound_ops import BoundConstant, BoundParams, BoundBuffers, BoundInput


def precompute_A(net, A, x, interm_bounds):
    # TODO maybe some can be pruned
    need_A = set()
    for nodes in net.split_activations.values():
        for node in nodes:
            for inp in node[0].inputs:
                if inp.name not in net.root_names and inp.perturbed:
                    need_A.add(inp)
    for node in need_A:
        if node.name not in A:
            print(f'Missing A for {node}. Making an additional CROWN call.')
            batch_size = node.output_shape[0]
            dim_output = int(np.prod(node.output_shape[1:]))
            C = torch.eye(dim_output, device=net.device).unsqueeze(0).expand(batch_size, -1, -1)
            ret = net.compute_bounds(
                x=(x,), C=C, method='CROWN', final_node_name=node.name,
                return_A=True,
                needed_A_dict={node.name: [net.input_name[0]]},
                interm_bounds=interm_bounds)
            A.update(ret[-1])


def set_roots(roots, x, A):
    assert isinstance(roots[0], BoundInput)
    assert isinstance(roots[0].value, BoundedTensor)
    roots[0].center = x
    roots[0].perturbation = x.ptb
    roots[0].aux = None
    roots[0].uA = None
    for r in roots[1:]:
        assert isinstance(r, (BoundConstant, BoundParams, BoundBuffers))
        if isinstance(r, BoundParams):
            assert isinstance(r.param, nn.Parameter)
            r.center = r.param
        r.perturbation = None
        r.lA = r.uA = None
    # (dim_output, batch_size, dim_input)
    roots[0].lA = A.sum(dim=2)



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import torch

class TensorStorage(object):
    """
    Fast managed dynamic sized tensor storage.
    """
    def __init__(self, full_shape, initial_size=1024, switching_size=65536,
                 dtype=None, device='cpu', concat_dim=0):
        """
        full_shape is the tensor shape you want to store using this object, including the "batch" dimension.
        dtype is tensor type (default float32).
        initial_size is the initial size of the storage. It will go up exponentially until reaching a batch size of "switching_size".
        switching_size is the point where exponential growth changes to linear growth.
        device is storage device, and for CPU memory it will be pinned.
        concat_dim is the axis of batch dimension (default is 0).
        """
        if isinstance(full_shape, torch.Tensor):
            data = full_shape
            full_shape = data.shape
        else:
            data = None
        if dtype is None:
            dtype = torch.get_default_dtype()
        self.shape = list(full_shape)  # Full shape, with batch size that will become dynamic.
        self.dtype = dtype
        self.device = device
        self.concat_dim = concat_dim
        self.num_used = 0
        self.switching_size = switching_size
        self._storage = self._allocate(initial_size)

        if data is not None:
            self.append(data)

    def _allocate(self, new_size):
        allocate_shape = self.shape.copy()
        allocate_shape[self.concat_dim] = new_size
        if self.device == 'cpu' and torch.cuda.is_available():
            # Pin CPU memory if cuda is available.
            return torch.empty(allocate_shape, dtype=self.dtype, device=self.device, pin_memory=True)
        else:
            return torch.empty(allocate_shape, dtype=self.dtype, device=self.device)

    def _get_new_size(self, request_size):
        """Compute new size of storage given the current request."""
        if self._storage.size(self.concat_dim) < self.switching_size:
            # Tensor is small. Exponential growth.
            return max(self._storage.size(self.concat_dim) * 2, self.num_used + request_size)
        else:
            # Tensor is big. Linear growth.
            return self._storage.size(self.concat_dim) + request_size * 32

    @torch.no_grad()
    def append(self, appended_tensor):
        """Append a new tensor to the storage object."""
        if self.num_used + appended_tensor.size(self.concat_dim) > self._storage.size(self.concat_dim):
            # Reallocate a new tensor, copying the existing contents over.
            new_size = self._get_new_size(appended_tensor.size(self.concat_dim))
            new_tensor = self._allocate(new_size)
            new_tensor.narrow(dim=self.concat_dim, start=0, length=self.num_used).copy_(
                self._storage.narrow(dim=self.concat_dim, start=0, length=self.num_used))
            # And then remove the old storage object.
            del self._storage
            self._storage = new_tensor
        self._storage.narrow(self.concat_dim, self.num_used, appended_tensor.size(self.concat_dim)).copy_(appended_tensor)
        self.num_used += appended_tensor.size(self.concat_dim)
        return self

    @torch.no_grad()
    def pop(self, size):
        """Remove tensors with 'size' at the end of the storage."""
        size = max(min(size, self.num_used), 0)
        ret = self._storage.narrow(self.concat_dim, self.num_used - size, size)
        self.num_used -= size
        return ret

    def tensor(self):
        return self._storage.narrow(self.concat_dim, 0, self.num_used)

    def __getattr__(self, attr):
        """Proxy all tensor attributes."""
        return getattr(self.tensor(), attr)

    def __getitem__(self, idx):
        return self.tensor()[idx]

    def __len__(self):
        return self.num_used

    def __sub__(self, o):
        return self._storage.narrow(self.concat_dim, 0, self.num_used) - o._storage.narrow(o.concat_dim, 0, o.num_used)


def _test():
    for concat_dim in [0, 1, 2]:
        shape = [1,1,1]
        shape[concat_dim] = -1 # does no matter.
        zero_shape = shape.copy()
        zero_shape[concat_dim] = 0
        make_tensor = lambda x: torch.arange(1,x+1, dtype=torch.float32).view(*shape)
        s = TensorStorage(full_shape=shape, initial_size=16, switching_size=65536, concat_dim=concat_dim)
        s.append(make_tensor(1))
        assert s.sum() == 1, print(s)
        s.append(make_tensor(3))
        assert s.sum() == 1 + 6, print(s)
        s.append(make_tensor(5))
        assert s.sum() == 1 + 6 + 15, print(s)
        t = s.pop(5)
        assert torch.allclose(t.squeeze(), torch.tensor([1,2,3,4,5], dtype=torch.float32)), print(t)
        t = s.pop(0)
        assert t.shape == torch.Size(zero_shape)
        t = s.pop(-1)
        assert t.shape == torch.Size(zero_shape)
        s.append(make_tensor(100))
        assert s.sum() == 1 + 6 + 50*101
        t = s.pop(5)
        assert torch.allclose(t.squeeze(), torch.tensor([96,97,98,99,100], dtype=torch.float32)), print(t)
        assert s.size(concat_dim) == 99, print(s.size())
        assert s._storage.size(concat_dim) == 104, print(s._storage.size())
        s.append(make_tensor(10))
        assert s.size(concat_dim) == 109, print(s.size())
        assert s._storage.size(concat_dim) == 208, print(s._storage.size())
        s.append(make_tensor(32768))
        assert s.size(concat_dim) == 32877, print(s.size())
        assert s._storage.size(concat_dim) == 32877, print(s._storage.size())
        s.pop(1)
        s.append(make_tensor(2))
        assert s.size(concat_dim) == 32878, print(s.size())
        assert s._storage.size(concat_dim) == 32877*2, print(s._storage.size())
        s.append(make_tensor(32800))
        s.append(make_tensor(100))
        assert s._storage.size(concat_dim) == 32877*2+100*32, print(s._storage.size())
        s.pop(100000)
        assert s._storage.size(concat_dim) == 32877*2+100*32, print(s._storage.size())
        assert s.size(concat_dim) == 0, print(s.size())
        t = s.pop(1)
        assert t.shape == torch.Size(zero_shape)
        t = s.pop(0)
        assert t.shape == torch.Size(zero_shape)
        t = s.pop(-1)
        assert t.shape == torch.Size(zero_shape)

if __name__ == "__main__":
    _test()
